# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
The EC2 deployer

This deployer will create and manage hosts for a topology using Amazon EC2.
"""

from cPickle import load
from boto.exception import BotoClientError, EC2ResponseError
from globus.provision.common.utils import create_ec2_connection 
from globus.provision.common.ssh import SSH, SSHCommandFailureException
from globus.provision.common.threads import MultiThread, GPThread, SIGINTWatcher
import random
import time
import sys
import traceback
import os.path
from globus.provision.common import log
from globus.provision.core.deploy import BaseDeployer, VM, ConfigureThread, WaitThread,\
    DeploymentException
from globus.provision.core.topology import DeployData, EC2DeployData, Node

class EC2VM(VM):
    """
    Represents a VM running on EC2.
    
    See the documentation on globus.provision.core.deploy.VM for details
    on what the VM class is used for.
    """
        
    def __init__(self, ec2_instance):
        self.ec2_instance = ec2_instance
        
    def __str__(self):
        return self.ec2_instance.id

class Deployer(BaseDeployer):
    """
    The EC2 deployer.
    """
  
    def __init__(self, *args, **kwargs):
        BaseDeployer.__init__(self, *args, **kwargs)
        self.conn = None
        self.instances = None
        self.vols = []
        self.supports_create_tags = True
        self.has_gp_sg = False

    def set_instance(self, inst):
        self.instance = inst
        self.__connect()         
    
    def __connect(self):
        config = self.instance.config
        
        try:
            log.debug("Connecting to EC2...")
            ec2_server_hostname = config.get("ec2-server-hostname")
            ec2_server_port = config.get("ec2-server-port")
            ec2_server_path = config.get("ec2-server-path")
            
            if ec2_server_hostname != None:
                self.conn = create_ec2_connection(ec2_server_hostname,
                                                  ec2_server_path,
                                                  ec2_server_port) 
            else:
                self.conn = create_ec2_connection()
            
            if self.conn == None:
                raise DeploymentException, "AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment variables are not set."

            log.debug("Connected to EC2.")
        except BotoClientError, exc:
            raise DeploymentException, "Could not connect to EC2. %s" % exc.reason
        
    def __get_security_groups(self, topology, node):
        sgs = topology.get_deploy_data(node, "ec2", "security_groups")
        if sgs is None:
            sgs = []

        if len(sgs) == 0:
            if self.has_gp_sg:
                sgs = ["globus-provision"]
            else:
                gp_sg = self.conn.get_all_security_groups(filters={"group-name":"globus-provision"})
                if len(gp_sg) == 0:
                    gp_sg = self.conn.create_security_group('globus-provision', 'Security group for Globus Provision instances')
                
                    # Internal
                    gp_sg.authorize(src_group = gp_sg)

                    # SSH
                    gp_sg.authorize('tcp', 22, 22, '0.0.0.0/0')
                
                    # GridFTP
                    gp_sg.authorize('tcp', 2811, 2811, '0.0.0.0/0')
                    gp_sg.authorize('udp', 2811, 2811, '0.0.0.0/0')
                    gp_sg.authorize('tcp', 50000, 51000, '0.0.0.0/0')
                
                    # MyProxy
                    gp_sg.authorize('tcp', 7512, 7512, '0.0.0.0/0')

                    # Galaxy
                    gp_sg.authorize('tcp', 8080, 8080, '0.0.0.0/0')
        
                sgs = ["globus-provision"]
                self.has_gp_sg = True
        else:
            all_sgs = self.conn.get_all_security_groups()
            # TODO: Validate that the security groups are valid
        
        return sgs
    
    def allocate_vm(self, node):
        topology = self.instance.topology
        
        instance_type = topology.get_deploy_data(node, "ec2", "instance_type")
        ami = topology.get_deploy_data(node, "ec2", "ami")
        security_groups = self.__get_security_groups(topology, node)

        try:
            image = self.conn.get_image(ami)
        except EC2ResponseError, ec2err:
            if ec2err.error_code in ("InvalidAMIID.NotFound", "InvalidAMIID.Malformed"):
                raise DeploymentException, "AMI %s does not exist" % ami
            else:
                raise ec2err

        if image == None:
            # Workaround for this bug:
            # https://bugs.launchpad.net/eucalyptus/+bug/495670
            image = [i for i in self.conn.get_all_images() if i.id == ami]
            if len(image) == 0:
                raise DeploymentException, "AMI %s does not exist" % ami
            else:
                image = image[0]         
        
        user_data = """#cloud-config
manage_etc_hosts: true
"""        
        
        log.info(" |- Launching a %s instance for %s." % (instance_type, node.id))
        reservation = image.run(min_count=1, 
                                max_count=1,
                                instance_type=instance_type,
                                security_groups= security_groups,
                                user_data = user_data,
                                key_name=self.instance.config.get("ec2-keypair"),
                                placement = None)
        instance = reservation.instances[0]
        
        return EC2VM(instance)

    def resume_vm(self, node):
        ec2_instance_id = node.deploy_data.ec2.instance_id

        log.info(" |- Resuming instance %s for %s." % (ec2_instance_id, node.id))        
        started = self.conn.start_instances([ec2_instance_id])            
        log.info(" |- Resumed instance %s." % ",".join([i.id for i in started]))
        
        return EC2VM(started[0])

    def post_allocate(self, node, vm):
        ec2_instance = vm.ec2_instance
        
        if ec2_instance.private_ip_address != None:
            # A correct EC2 system should return this
            node.ip = ec2_instance.private_ip_address
        else:
            # Unfortunately, some EC2-ish systems won't return the private IP address
            # We fall back on the private_dns_name, which should still work
            # (plus, some EC2-ish systems actually set this to the IP address)
            node.ip = ec2_instance.private_dns_name

        node.hostname = ec2_instance.public_dns_name

        # TODO: The following won't work on EC2-ish systems behind a firewall.
        node.public_ip = ".".join(ec2_instance.public_dns_name.split(".")[0].split("-")[1:])

        if not node.has_property("deploy_data"):
            node.deploy_data = DeployData()
            node.deploy_data.ec2 = EC2DeployData()            

        node.deploy_data.ec2.instance_id = ec2_instance.id

        try:
            if self.supports_create_tags:
                self.conn.create_tags([ec2_instance.id], {"Name": "%s_%s" % (self.instance.id, node.id)})
        except:
            # Some EC2-ish systems don't support the create_tags call.
            # If it fails, we just silently ignore it, as it is not essential,
            # but make sure not to call it again, as EC2-ish systems will
            # timeout instead of immediately returning an error
            self.supports_create_tags = False


    def get_node_vm(self, nodes):
        ec2_instance_ids = [n.deploy_data.ec2.instance_id for n in nodes]
        reservations = self.conn.get_all_instances(ec2_instance_ids)
        node_vm = {}
        for r in reservations:
            instance = r.instances[0]
            node = [n for n in nodes if n.deploy_data.ec2.instance_id==instance.id][0]
            node_vm[node] = EC2VM(instance)
        return node_vm

    def stop_vms(self, nodes):
        ec2_instance_ids = [n.deploy_data.ec2.instance_id for n in nodes]
        log.info("Stopping EC2 instances %s." % ", ".join(ec2_instance_ids))
        stopped = self.conn.stop_instances(ec2_instance_ids)
        log.info("Stopped EC2 instances %s." % ", ".join([i.id for i in stopped]))

    def terminate_vms(self, nodes):
        ec2_instance_ids = [n.deploy_data.ec2.instance_id for n in nodes]
        log.info("Terminating EC2 instances %s." % ", ".join(ec2_instance_ids))
        terminated = self.conn.terminate_instances(ec2_instance_ids)
        log.info("Terminated EC2 instances %s." % ", ".join([i.id for i in terminated]))
        
    def wait_state(self, obj, state, interval = 2.0):
        jitter = random.uniform(0.0, 0.5)
        while True:
            time.sleep(interval + jitter)
            try:
                newstate = obj.update()
            except EC2ResponseError, ec2err:
                if ec2err.error_code == "InvalidInstanceID.NotFound":
                    # If the instance was just created, this is a transient error. 
                    # We just have to wait until the instance appears.
                    pass
                else:
                    raise ec2err            
            
            if newstate == state:
                return True
        # TODO: Check errors            
        
    def get_wait_thread_class(self):
        return self.NodeWaitThread

    def get_configure_thread_class(self):
        return self.NodeConfigureThread
            
    class NodeWaitThread(WaitThread):
        def __init__(self, multi, name, node, vm, deployer, state, depends = None):
            WaitThread.__init__(self, multi, name, node, vm, deployer, state, depends)
            self.ec2_instance = vm.ec2_instance
                        
        def wait(self):
            if self.state in (Node.STATE_RUNNING_UNCONFIGURED, Node.STATE_RESUMED_UNCONFIGURED):
                self.deployer.wait_state(self.ec2_instance, "running")
                log.info("Instance %s is running. Hostname: %s" % (self.ec2_instance.id, self.ec2_instance.public_dns_name))
            elif self.state == Node.STATE_STOPPED:
                self.deployer.wait_state(self.ec2_instance, "stopped")
            elif self.state == Node.STATE_TERMINATED:
                self.deployer.wait_state(self.ec2_instance, "terminated")
            
            
    class NodeConfigureThread(ConfigureThread):
        def __init__(self, multi, name, node, vm, deployer, depends = None, basic = True, chef = True):
            ConfigureThread.__init__(self, multi, name, node, vm, deployer, depends, basic, chef)
            self.ec2_instance = self.vm.ec2_instance
            
        def connect(self):
            keyfile = os.path.expanduser(self.config.get("ec2-keyfile"))
            ssh = self.ssh_connect(self.config.get("ec2-username"), self.ec2_instance.public_dns_name, keyfile)
            return ssh
        
        def pre_configure(self, ssh):
            node = self.node
            instance = self.ec2_instance
            
            log.info("Setting up instance %s. Hostname: %s" % (instance.id, instance.public_dns_name), node)
           
            try:
                ssh.run("ls -l /chef")
            except SSHCommandFailureException:
                #The image is not properly setup, so do all pre-configuration for globus-provision
                log.info("Image is not configured with Chef, so installing...")

                ssh.run("sudo chown -R %s /chef" % self.config.get("ec2-username"))
                ssh.scp_dir("%s" % self.chef_dir, "/chef")



                ssh.run("addgroup admin", exception_on_error = False)
                ssh.run("echo \"%s `hostname`\" | sudo tee -a /etc/hosts" % instance.private_ip_address)

                ssh.run("sudo apt-get install lsb-release wget")
                ssh.run("echo \"deb http://apt.opscode.com/ `lsb_release -cs` main\" | sudo tee /etc/apt/sources.list.d/opscode.list")
                ssh.run("wget -qO - http://apt.opscode.com/packages@opscode.com.gpg.key | sudo apt-key add -")
                ssh.run("sudo apt-get update")
                ssh.run("echo 'chef chef/chef_server_url string http://127.0.0.1:4000' | sudo debconf-set-selections")
                ssh.run("sudo apt-get -q=2 install chef")
        
                ssh.run("echo -e \"cookbook_path \\\"/chef/cookbooks\\\"\\nrole_path \\\"/chef/roles\\\"\" > /tmp/chef.conf")        
                ssh.run("echo '{ \"run_list\": \"recipe[provision::ec2]\", \"scratch_dir\": \"%s\" }' > /tmp/chef.json" % self.scratch_dir)

                ssh.run("sudo chef-solo -c /tmp/chef.conf -j /tmp/chef.json")    
        
                ssh.run("sudo update-rc.d -f nis remove")
                ssh.run("sudo update-rc.d -f condor remove")
                ssh.run("sudo update-rc.d -f chef-client remove")
       

                log.debug("Removing private data...")
         
                ssh.run("sudo find /root/.*history /home/*/.*history -exec rm -f {} \;", exception_on_error = False)


            
                
        def post_configure(self, ssh):
            pass
            



# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Core deployment classes.

These classes contain code that is common to all deployers (i.e., they don't 
contain any infrastructure-specific code).

To create a new deployer, you will need to extend classes the classes
in this module.

"""

from globus.provision.common.threads import GPThread
from globus.provision.common.ssh import SSH, SSHCommandFailureException
from globus.provision.common import log
from globus.provision.core.topology import Node

from abc import ABCMeta, abstractmethod

class DeploymentException(Exception):
    """A simple exception class used for deployment exceptions"""
    pass

class BaseDeployer(object):
    """
    The base class for a deployer.
    
    A deployer must implement all the abstract methods in this class
    """
    
    __metaclass__ = ABCMeta
    
    def __init__(self, extra_files = [], run_cmds = []):
        self.instance = None
        self.extra_files = extra_files
        self.run_cmds = run_cmds
        
    @abstractmethod
    def set_instance(self, inst): pass
    
    @abstractmethod
    def allocate_vm(self, node): pass

    @abstractmethod
    def post_allocate(self, node, vm): pass

    @abstractmethod
    def stop_vms(self, nodes): pass
    
    @abstractmethod
    def resume_vm(self, node): pass
        
    @abstractmethod
    def terminate_vms(self, nodes): pass
        
    @abstractmethod
    def get_node_vm(self, nodes): pass
        
    @abstractmethod
    def get_wait_thread_class(self): pass

    @abstractmethod
    def get_configure_thread_class(self): pass

            
class VM(object):
    """
    A VM object represents a virtual machine managed by a
    deployer. It is basically meant as an opaque type that
    can be returned by the deployer to the core, and then
    passed from the core to other functions in the deployer.
    """
    def __init__(self):
        pass
    
class WaitThread(GPThread):
    """
    The base class for "waiter threads".
    
    A derived class must implement the wait() method, with the
    deployer-specific code that will wait until a VM
    has reached a given state.
    """
    
    __metaclass__ = ABCMeta
        
    def __init__(self, multi, name, node, vm, deployer, state, depends):
        GPThread.__init__(self, multi, name, depends)
        self.node = node
        self.vm = vm
        self.deployer = deployer
        self.state = state

    def run2(self):
        topology = self.deployer.instance.topology
        
        self.wait()
        
        self.node.state = self.state
        topology.save()
        
    @abstractmethod
    def wait(self): pass
         

class ConfigureThread(GPThread):
    """
    The base class for "configure threads".
    
    This is a thread that takes care of configuring a single VM.
    Most of the actions (e.g., SSH'ing to the VM and running Chef)
    will be the same in most deployers. So, this class simply
    requires that derived classes implement pre_configure()
    and post_configure(), in case there are deployer-specific
    actions that must be taken. The connect() method must
    also be implemented, although it can usually just be
    a call to ssh_connect.
    """    
    
    __metaclass__ = ABCMeta
        
    def __init__(self, multi, name, node, vm, deployer, depends = None, basic = True, chef = True, dryrun=False):
        GPThread.__init__(self, multi, name, depends)
        self.domain = node.parent_Domain
        self.node = node
        self.vm = vm
        self.deployer = deployer
        self.config = deployer.instance.config
        self.basic = basic
        self.chef = chef
        self.dryrun = dryrun

    def run2(self):
        topology = self.deployer.instance.topology
        
        if self.node.state in (Node.STATE_RUNNING_UNCONFIGURED, Node.STATE_RUNNING, Node.STATE_RESUMED_UNCONFIGURED):
            if self.node.state == Node.STATE_RUNNING_UNCONFIGURED:
                log.debug("Configuring node for the first time", self.node)
                self.node.state = Node.STATE_CONFIGURING
                next_state = Node.STATE_RUNNING
            elif self.node.state == Node.STATE_RUNNING:
                log.debug("Reconfiguring already-running node", self.node)
                self.node.state = Node.STATE_RECONFIGURING
                next_state = Node.STATE_RUNNING
            elif self.node.state == Node.STATE_RESUMED_UNCONFIGURED:
                log.debug("Reconfiguring resumed node", self.node)
                self.node.state = Node.STATE_RESUMED_RECONFIGURING
                next_state = Node.STATE_RUNNING
            
            topology.save()
            
            if not self.dryrun:
                ssh = self.connect()
                self.check_continue()
                self.pre_configure(ssh)
                self.check_continue()
                self.configure(ssh)
                self.check_continue()
                self.post_configure(ssh)
                self.check_continue()
    
            self.node.state = next_state
            topology.save()
        elif self.node.state == Node.STATE_STOPPING:
            log.debug("Doing pre-shutdown configuration", self.node)
            self.node.state = Node.STATE_STOPPING_CONFIGURING
            topology.save()
            
            if not self.dryrun:
                ssh = self.connect()
                self.check_continue()
                self.configure_stop(ssh)
                self.check_continue()
    
            self.node.state = Node.STATE_STOPPING_CONFIGURED
            topology.save()            

    @abstractmethod
    def connect(self): pass

    @abstractmethod
    def pre_configure(self): pass

    @abstractmethod
    def post_configure(self): pass
        
    def ssh_connect(self, username, hostname, keyfile):
        node = self.node

        log.debug("Establishing SSH connection", node)
        ssh = SSH(username, hostname, keyfile, default_outf = None, default_errf = None)
        try:
            ssh.open()
        except Exception, e:
            log.debug("SSH connection timed out", node)
            # Raise exception and let multi-thread manager handle it
            raise e
        log.debug("SSH connection established", node)
        
        return ssh
        
    def configure(self, ssh):
        domain = self.domain
        node = self.node
        instance_dir = self.deployer.instance.instance_dir        
        
        if self.basic:
            # Make backup copies of hostname and /etc/hosts
            if node.state in (Node.STATE_CONFIGURING, Node.STATE_RESUMED_RECONFIGURING):
                ssh.run("sudo cp /etc/hosts /etc/hosts.gp-bak", expectnooutput=True)
                ssh.run("sudo cp /etc/hostname /etc/hostname.gp-bak", expectnooutput=True)
            
            # Upload host file and update hostname
            log.debug("Uploading host file and updating hostname", node)
            ssh.scp("%s/hosts" % instance_dir,
                    "/chef/cookbooks/provision/files/default/hosts")             
            ssh.run("sudo cp /chef/cookbooks/provision/files/default/hosts /etc/hosts", expectnooutput=True)
    
            ssh.run("sudo bash -c \"echo %s > /etc/hostname\"" % node.hostname, expectnooutput=True)
            ssh.run("sudo /etc/init.d/hostname.sh || sudo /etc/init.d/hostname restart", expectnooutput=True)
        
        self.check_continue()

        if self.chef:        
            # Upload topology file
            log.debug("Uploading topology file", node)
            ssh.scp("%s/topology.rb" % instance_dir,
                    "/chef/cookbooks/provision/attributes/topology.rb")             
            
            # Copy certificates
            log.debug("Copying certificates", node)
            ssh.scp_dir("%s/certs" % instance_dir, 
                        "/chef/cookbooks/provision/files/default/")
    
            # Upload extra files
            log.debug("Copying extra files", node)
            for src, dst in self.deployer.extra_files:
                ssh.scp(src, dst)
            
            self.check_continue()

            # Run chef
            log.debug("Running chef", node)
            ssh.run("echo -e \"cookbook_path \\\"/chef/cookbooks\\\"\\nrole_path \\\"/chef/roles\\\"\" > /tmp/chef.conf", expectnooutput=True)        
            ssh.run("echo '{ \"run_list\": [ %s ], \"scratch_dir\": \"%s\", \"domain_id\": \"%s\", \"node_id\": \"%s\"  }' > /tmp/chef.json" % (",".join("\"%s\"" % r for r in node.run_list), self.config.get("scratch-dir"), domain.id, node.id), expectnooutput=True)
            
            # Sometimes, Chef will fail because a service didn't start or restart
            # properly (NFS-related services seem to do this occasionally).
            # In most cases, the problem just "goes away" if you try to restart the
            # service again. So, if Chef fails, we don't give up and try again
            # (since the recipes are idempotent, there's no harm to running them
            # multiple times)
            chef_tries = 3
            while chef_tries > 0:
                rc = ssh.run("sudo -i chef-solo -c /tmp/chef.conf -j /tmp/chef.json", exception_on_error = False)    
                if rc != 0:
                    chef_tries -= 1
                    log.debug("chef-solo failed. %i attempts left", node)
                else:
                    break
                    
            if chef_tries == 0:
                raise DeploymentException, "Failed to configure node %s" % node.id
                    
            self.check_continue() 

        for cmd in self.deployer.run_cmds:
            ssh.run(cmd)

        log.info("Configuration done.", node)
        
    def configure_stop(self, ssh):      
        node = self.node
  
        log.info("Configuring node for shutdown", node)
        ssh.run("sudo cp /etc/hosts.gp-bak /etc/hosts", expectnooutput=True)
        ssh.run("sudo cp /etc/hostname.gp-bak /etc/hostname", expectnooutput=True)
        ssh.run("sudo /etc/init.d/hostname.sh || sudo /etc/init.d/hostname restart", expectnooutput=True)
        ssh.run("sudo bash -c \"echo +auto.master > /etc/auto.master\"", exception_on_error = False)
        ssh.run("sudo bash -c \"echo > /etc/yp.conf\"", exception_on_error = False)
        ssh.run("sudo bash -c \"echo > /etc/default/nfs-common\"", exception_on_error = False)

        ssh.run("sudo update-rc.d -f nis remove", exception_on_error = False)
        log.info("Configuration done.", node)
        


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Certificate management
"""

from OpenSSL import crypto
import os.path

class CertificateGenerator(object):
    """
    A class for generating CA, host, and user certificates.
    """
    
    def __init__(self, dn, ca_cert = None, ca_key = None):
        self.ca_cert = ca_cert
        self.ca_key = ca_key
        self.serial = 1

        self.dn = [x.split("=") for x in dn.split(",")]
        self.dn = dict([(n.upper().strip(), v.strip()) for n,v in self.dn])
        
    def set_ca(self, ca_cert, ca_key):
        self.ca_cert = ca_cert
        self.ca_key = ca_key
        
    def gen_selfsigned_ca_cert(self, cn):
        return self.__gen_certificate(cn = cn, 
                                      issuer_cert = None,
                                      issuer_key = None)

    def gen_user_cert(self, cn):
        return self.__gen_certificate(cn = cn, 
                                      issuer_cert = self.ca_cert,
                                      issuer_key = self.ca_key)

    def gen_host_cert(self, hostname):
        return self.__gen_certificate(cn = "host/"+hostname, 
                                      issuer_cert = self.ca_cert,
                                      issuer_key = self.ca_key)
        
    def __gen_certificate(self, cn, issuer_cert, issuer_key):
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 1024)

        cert = crypto.X509()
        if self.dn.has_key("C"):
            cert.get_subject().C = self.dn["C"]
        if self.dn.has_key("ST"):
            cert.get_subject().ST = self.dn["ST"]
        if self.dn.has_key("L"):
            cert.get_subject().L = self.dn["L"]
        if self.dn.has_key("O"):
            cert.get_subject().O = self.dn["O"]
        if self.dn.has_key("OU"):
            cert.get_subject().OU = self.dn["OU"]
        cert.get_subject().CN = cn
        cert.set_serial_number(self.serial)
        self.serial += 1
        cert.set_version(2)
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(10*365*24*60*60)
        cert.set_pubkey(k)

        
        if issuer_cert == None:
            cert.set_issuer(cert.get_subject())
        else:
            cert.set_issuer(issuer_cert.get_subject())
            
        if issuer_cert == None:
            cert.sign(k, 'sha1')
        else:
            cert.sign(issuer_key, 'sha1')   
        
        return cert, k 
    
    # TODO: Return something meaningful depending on whether the
    # certificate was regenerated or not
    def save_certificate(self, cert, key, cert_file, key_file, force = False):
        if not os.path.exists(cert_file) or force:
            cert_file = open(cert_file, "w")
            cert_file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
            cert_file.close()  
        
            key_file = open(key_file, "w")
            key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key))
            key_file.close()        
        
    def load_certificate(self, cert_file, key_file):
        cert_file = open(cert_file, "r")
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_file.read())
        cert_file.close()  
    
        key_file = open(key_file, "r")
        key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_file.read())
        key_file.close()      
        
        return cert, key


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
The dummy deployer

All the actions in this deployer simply return immediately, simulating a
backend that never fails. This deployer is useful for testing.
"""

from globus.provision.common.threads import GPThread
import sys
from globus.provision.common import log
from globus.provision.core.deploy import BaseDeployer, VM, ConfigureThread, WaitThread
from globus.provision.core.topology import Node

class DummyVM(VM):
    """
    A "dummy VM". Doesn't actually contain anything.
    
    See the documentation on globus.provision.core.deploy.VM for details
    on what the VM class is used for.
    """
    
    def __init__(self):
        VM.__init__(self)
        
class Deployer(BaseDeployer):
    """
    The dummy deployer
    """
  
    def __init__(self, *args, **kwargs):
        BaseDeployer.__init__(self, *args, **kwargs)

    def set_instance(self, inst):
        self.instance = inst
        
    def allocate_vm(self, node):
        log.info("Allocated dummy VM.")     
        return DummyVM()

    def resume_vm(self, node):
        log.info("Resumed dummy VM.")     
        return DummyVM()

    def post_allocate(self, node, vm):
        node.hostname = "%s.gp.example.org" % node.id
        node.ip = "1.2.3.4"

    def get_node_vm(self, nodes):
        node_vm = {}
        for n in nodes:
            node_vm[n] = DummyVM()
        return node_vm

    def stop_vms(self, nodes):
        log.info("Dummy nodes terminated.")         

    def terminate_vms(self, nodes):
        log.info("Dummy nodes terminated.")

    def get_wait_thread_class(self):
        return self.NodeWaitThread

    def get_configure_thread_class(self):
        return self.NodeConfigureThread
            
    class NodeWaitThread(WaitThread):
        def __init__(self, multi, name, node, vm, deployer, state, depends = None):
            WaitThread.__init__(self, multi, name, node, vm, deployer, state, depends)
                        
        def wait(self):
            log.info("Waiting for state %s" % Node.state_str[self.state])
            
    class NodeConfigureThread(ConfigureThread):
        def __init__(self, multi, name, node, vm, deployer, depends = None, basic = True, chef = True):
            ConfigureThread.__init__(self, multi, name, node, vm, deployer, depends, basic, chef, dryrun = True)            

        def connect(self): pass
    
        def pre_configure(self): pass
    
        def post_configure(self): pass
            


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Convenience functions around the logging module
"""

import logging
import os.path

def init_logging(level):
    if level == 2:
        level = logging.DEBUG
    elif level == 1:
        level = logging.INFO
    else:
        level = logging.WARNING
    logging.getLogger('boto').setLevel(logging.CRITICAL)
    logging.getLogger('paramiko').setLevel(logging.CRITICAL)
    
    l = logging.getLogger("globusprovision")
    l.setLevel(logging.DEBUG)
    
    fh = logging.StreamHandler()
    fh.setLevel(level)
    formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s')
    fh.setFormatter(formatter)
    l.addHandler(fh)        

def set_logging_instance(instance):
    l = logging.getLogger("globusprovision")
    fh = logging.FileHandler(os.path.expanduser('%s/deploy.log' % instance.instance_dir))
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(message)s')
    fh.setFormatter(formatter)
    l.addHandler(fh)    

def log(msg, func, node):
    if node != None:
        msg = "%s - %s" % (node.id, msg)
    func(msg)

def debug(msg, node = None):
    log(msg, logging.getLogger('globusprovision').debug, node)

def warning(msg, node = None):
    log(msg, logging.getLogger('globusprovision').warning, node)
    
def info(msg, node = None):
    log(msg, logging.getLogger('globusprovision').info, node)



# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #
import json

"""
Commands that directly invoke the API. Most of these commands are a one-to-one
mapping to the API, although some some (like gp-instance-add-host) simply provide 
a more convenient interface on top of instance_update().
"""

import time
import sys

from colorama import Fore, Style

import globus.provision.common.defaults as defaults

from globus.provision.cli import Command
from globus.provision.core.api import API
from globus.provision.common.utils import parse_extra_files_files
from globus.provision.common.threads import SIGINTWatcher
from globus.provision.core.topology import Topology, Node, User
from globus.provision.core.config import SimpleTopologyConfig
from globus.provision.common.config import ConfigException


def gp_instance_create_func():
    return gp_instance_create(sys.argv).run()
        
class gp_instance_create(Command):
    """
    Creates a new Globus Provision instance.
    """
    
    name = "gp-instance-create"

    def __init__(self, argv):
        Command.__init__(self, argv, disable_sigintwatch=True)
        
        self.optparser.add_option("-c", "--conf", 
                                  action="store", type="string", dest="conf", 
                                  default = defaults.CONFIG_FILE,
                                  help = "Configuration file.")
        
        self.optparser.add_option("-t", "--topology", 
                                  action="store", type="string", dest="topology",
                                  help = "Topology file. Can be either a simple topology file (with extension .conf) "
                                         "or a topology JSON file (with extension .json).")        
                
    def run(self):    
        self.parse_options()

        if self.opt.conf is None:
            print "You must specify a configuration file using the -c/--conf option."
            return 1

        self._check_exists_file(self.opt.conf)

        if self.opt.topology is None:
            topology_file = self.opt.conf
        else:
            self._check_exists_file(self.opt.topology)
            topology_file = self.opt.topology
        
        if topology_file.endswith(".json"):
            jsonfile = open(topology_file)
            topology_json = jsonfile.read()
            jsonfile.close()
        elif topology_file.endswith(".conf"):
            try:
                conf = SimpleTopologyConfig(topology_file)
                topology = conf.to_topology()
                topology_json = topology.to_json_string()
            except ConfigException, cfge:
                self._print_error("Error in topology configuration file.", cfge)
                return 1         
        else:   
            self._print_error("Unrecognized topology file format.", "File must be either a JSON (.json) or configuration (.conf) file.")
            return 1         

        configf = open(self.opt.conf)
        config_txt = configf.read()
        configf.close()

        api = API(self.opt.dir)
        (status_code, message, inst_id) = api.instance_create(topology_json, config_txt)

        if status_code != API.STATUS_SUCCESS:
            self._print_error("Could not create instance.", message)
            return 1
        else:
            print "Created new instance:",
            print Fore.WHITE + Style.BRIGHT + inst_id
            self._set_last_gpi(inst_id)
            return 0
            
def gp_instance_describe_func():
    return gp_instance_describe(sys.argv).run()            
        
class gp_instance_describe(Command):
    """
    Describes a Globus Provision instance, providing information on the state of the instance,
    and of the individual hosts (including their hostnames and IPs, if the instance is running). 
    Running in verbose mode (with the ``-v`` option)
    will print out the raw JSON representation of the instance's topology.
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-instance-describe --verbose gpi-12345678
        
    """
    
    name = "gp-instance-describe"
    
    def __init__(self, argv):
        Command.__init__(self, argv, disable_sigintwatch=True)
                
    def run(self):    
        self.parse_options()
        
        if len(self.args) != 2:
            print "You must specify an instance id."
            print "For example: %s [options] gpi-37a8bf17" % self.name
            return 1
         
        inst_id = self.args[1]
        
        api = API(self.opt.dir)
        (status_code, message, topology_json) = api.instance(inst_id)
        
        if status_code != API.STATUS_SUCCESS:
            self._print_error("Could not access instance.", message)
            return 1 
        else:
            self._set_last_gpi(inst_id)
            
            if self.opt.verbose or self.opt.debug:
                print topology_json
            else:
                topology = Topology.from_json_string(topology_json)
                reset = Fore.RESET + Style.RESET_ALL
                print Fore.WHITE + Style.BRIGHT + inst_id + reset + ": " + self._colorize_topology_state(topology.state)
                print
                for domain in topology.domains.values():
                    print "Domain " + Fore.CYAN + "'%s'" % domain.id
                    
                    # Find "column" widths and get values while we're at it
                    node_width = state_width = hostname_width = ip_width = 0
                    rows = []
                    for node in domain.get_nodes():
                        if len(node.id) > node_width:
                            node_width = len(node.id)                        
                        
                        if node.has_property("state"):
                            state = node.state
                        else:
                            state = Node.STATE_NEW                            
                        state_str = Node.state_str[state]

                        if len(state_str) > state_width:
                            state_width = len(state_str)

                        if node.has_property("hostname"):
                            hostname = node.hostname
                        else:
                            hostname = ""                            

                        if len(hostname) > hostname_width:
                            hostname_width = len(hostname)

                        if node.has_property("ip"):
                            ip = node.ip
                        else:
                            ip = "" 

                        if len(ip) > ip_width:
                            ip_width = len(ip)    
                            
                        rows.append((node.id, state, state_str, hostname, ip))                          
                                                
                    for (node_id, state, state_str, hostname, ip) in rows:
                        node_id_pad = self._pad(node_id, Fore.WHITE + Style.BRIGHT + node_id + Fore.RESET + Style.RESET_ALL, node_width + 2) 
                        state_pad = self._pad(state_str, self._colorize_node_state(state), state_width + 2) 
                        hostname_pad   = self._pad(hostname, "", hostname_width + 2)
                        ip_pad = self._pad(ip, "", ip_width)
                        print "    " + node_id_pad + state_pad + hostname_pad + ip_pad
                    print
            
            return 0
                    
def gp_instance_start_func():
    return gp_instance_start(sys.argv).run()  

class gp_instance_start(Command):
    """
    Starts a Globus Provision instance. If the instance was previous stopped, ``gp-instance-start``
    will resume it.
    
    See :ref:`sec_test_chef` for details on how to use the ``--extra-files`` option.
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-instance-start --extra-files foo.txt gpi-12345678
        
    """
        
    name = "gp-instance-start"
    
    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)

        self.optparser.add_option("-x", "--extra-files", 
                                  action="store", type="string", dest="extra_files", 
                                  help = "File with list of files to upload to each host before configuring the instance.")

        self.optparser.add_option("-r", "--run", 
                                  action="store", type="string", dest="run", 
                                  help = "File with list of commands to run on each host after configuring the instance.")
                        
    def run(self):        
        t_start = time.time()        
        self.parse_options()
        
        if len(self.args) != 2:
            print "You must specify an instance id."
            print "For example: %s [options] gpi-37a8bf17" % self.name
            return 1
        
        inst_id = self.args[1]
            
        if self.opt.extra_files != None:
            self._check_exists_file(self.opt.extra_files)
            extra_files = parse_extra_files_files(self.opt.extra_files)
        else:
            extra_files = []

        if self.opt.run != None:
            self._check_exists_file(self.opt.run)
            run_cmds = [l.strip() for l in open(self.opt.run).readlines()]
        else:
            run_cmds = []

        print "Starting instance",
        print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",
        api = API(self.opt.dir)
        status_code, message = api.instance_start(inst_id, extra_files, run_cmds)
        
        if status_code == API.STATUS_SUCCESS:
            print Fore.GREEN + Style.BRIGHT + "done!"

            self._set_last_gpi(inst_id)
            
            t_end = time.time()
            
            delta = t_end - t_start
            minutes = int(delta / 60)
            seconds = int(delta - (minutes * 60))
            print "Started instance in " + Fore.WHITE + Style.BRIGHT + "%i minutes and %s seconds" % (minutes, seconds)
            return 0
        elif status_code == API.STATUS_FAIL:
            print
            self._print_error("Could not start instance.", message)
            return 1 

def gp_instance_update_func():
    return gp_instance_update(sys.argv).run()  

class gp_instance_update(Command):
    """
    Updates a Globus Provision instance's topology. Globus Provision will determine what changes
    and necessary (adding/removing hosts, etc.) and will return an error if an invalid update
    was specified.
    
    See :ref:`sec_test_chef` for details on how to use the ``--extra-files`` option.
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-instance-update --topology newtopology.json gpi-12345678
        
    """
        
    name = "gp-instance-update"

    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)

        self.optparser.add_option("-t", "--topology", 
                                  action="store", type="string", dest="topology",
                                  help = "Topology file (JSON format only)")

        self.optparser.add_option("-x", "--extra-files", 
                                  action="store", type="string", dest="extra_files", 
                                  help = "File with list of files to upload to each host before configuring the instance.")
        
        self.optparser.add_option("-r", "--run", 
                                  action="store", type="string", dest="run", 
                                  help = "File with list of commands to run on each host after configuring the instance.")        
                
    def run(self):                
        t_start = time.time()        
        self.parse_options()

        if len(self.args) != 2:
            print "You must specify an instance id."
            print "For example: %s [options] gpi-37a8bf17" % self.name
            return 1
        
        inst_id = self.args[1]

        if self.opt.topology != None:
            self._check_exists_file(self.opt.topology)
    
            jsonfile = open(self.opt.topology)
            topology_json = jsonfile.read()
            jsonfile.close()
        else:
            topology_json = None
                

        if self.opt.extra_files != None:
            self._check_exists_file(self.opt.extra_files)
            extra_files = parse_extra_files_files(self.opt.extra_files)
        else:
            extra_files = []
            
        if self.opt.run != None:
            self._check_exists_file(self.opt.run)
            run_cmds = [l.strip() for l in open(self.opt.run).readlines()]
        else:
            run_cmds = []            

        print "Updating topology of",
        print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",

        api = API(self.opt.dir)
        status_code, message = api.instance_update(inst_id, topology_json, extra_files, run_cmds)
        
        if status_code == API.STATUS_SUCCESS:
            print Fore.GREEN + Style.BRIGHT + "done!"
            
            self._set_last_gpi(inst_id)
            
            t_end = time.time()
            
            delta = t_end - t_start
            minutes = int(delta / 60)
            seconds = int(delta - (minutes * 60))
            print "Updated topology in " + Fore.WHITE + Style.BRIGHT + "%i minutes and %s seconds" % (minutes, seconds)
            return 0
        elif status_code == API.STATUS_FAIL:
            self._print_error("Could not update topology.", message)
            return 1


def gp_instance_stop_func():
    return gp_instance_stop(sys.argv).run()         
        
class gp_instance_stop(Command):
    """
    Stops a running Globus Provision instance. This will shut down all the hosts in the instance,
    but it will not free the corresponding resources. You can use :ref:`cli_gp-instance-start` to resume
    the instance at a later time. Use :ref:`cli_gp-instance-terminate` if you want to shut down the hosts
    *and* free all their resources (including all associated storage) 
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-instance-stop --verbose gpi-12345678
        
    """
        
    name = "gp-instance-stop"
    
    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)
                
    def run(self):       
        self.parse_options()
        
        if len(self.args) != 2:
            print "You must specify an instance id."
            print "For example: %s [options] gpi-37a8bf17" % self.name
            return 1
        
        inst_id = self.args[1]            

        print "Stopping instance",
        print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",
        api = API(self.opt.dir)
        status_code, message = api.instance_stop(inst_id)
        
        if status_code == API.STATUS_SUCCESS:
            print Fore.GREEN + Style.BRIGHT + "done!"
            self._set_last_gpi(inst_id)
            return 0
        elif status_code == API.STATUS_FAIL:
            self._print_error("Could not stop instance.", message)
            print
            return 1 


def gp_instance_terminate_func():
    return gp_instance_terminate(sys.argv).run()     

class gp_instance_terminate(Command):
    """
    Terminates a Globus Provision instance. This not only shuts down all the hosts in the
    instance, but also frees up all associated resources, including storage. Use this command
    only if you want to irreversibly "kill" your instance. If you only want to stop it temporarily
    (shutting down the hosts, but allowing them to be resumed at a later time), use 
    :ref:`cli_gp-instance-stop` instead.
    
    This command can also be used on instances that are in the "Failed" state, to free their
    resources.
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-instance-terminate --verbose gpi-12345678
        
    """
    
    name = "gp-instance-terminate"
    
    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)
                
    def run(self):        
        self.parse_options()

        if len(self.args) != 2:
            print "You must specify an instance id."
            print "For example: %s [options] gpi-37a8bf17" % self.name
            return 1
                
        inst_id = self.args[1]

        print "Terminating instance",
        print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",
        api = API(self.opt.dir)
        status_code, message = api.instance_terminate(inst_id)
        
        if status_code == API.STATUS_SUCCESS:
            print Fore.GREEN + Style.BRIGHT + "done!"
            self._set_last_gpi(inst_id)
            return 0
        elif status_code == API.STATUS_FAIL:
            print
            self._print_error("Could not terminate instance.", message)
            return 1  


def gp_instance_list_func():
    return gp_instance_list(sys.argv).run()     

class gp_instance_list(Command):
    """
    Lists all instances you've created, showing their identifier and their state.
    
    If you only want to list certain instances, you can specify a list of instance
    identifiers after all other parameters. For example::
    
        gp-instance-list --verbose gpi-11111111 gpi-22222222 gpi-33333333
        
    """
        
    name = "gp-instance-list"
    
    def __init__(self, argv):
        Command.__init__(self, argv, disable_sigintwatch=True)
                
    def run(self):    
        self.parse_options()
        
        if len(self.args) >= 2:
            inst_ids = self.args[1:]
        else:
            inst_ids = None
        
        api = API(self.opt.dir)
        (status_code, message, topologies_json) = api.instance_list(inst_ids)
        
        
        if status_code != API.STATUS_SUCCESS:
            self._print_error("Unable to list instances.", message)
            return 1
        else:        
            insts = json.loads(topologies_json)
            
            for inst in insts:
                if self.opt.verbose or self.opt.debug:
                    print json.dumps(inst, indent=2)
                else:
                    topology = Topology.from_json_dict(inst)
                    
                    reset = Fore.RESET + Style.RESET_ALL
                    print Fore.WHITE + Style.BRIGHT + topology.id + reset + ": " + self._colorize_topology_state(topology.state)
            
            return 0

def gp_instance_add_user_func():
    return gp_instance_add_user(sys.argv).run()     

class gp_instance_add_user(Command):
    """
    Adds a user to a running instance.
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-instance-add-user --domain simple --login johnsmith gpi-12345678
        
    """
        
    name = "gp-instance-add-user"
    
    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)

        self.optparser.add_option("-m", "--domain", 
                                  action="store", type="string", dest="domain",
                                  help = "Domain that the user will be added to.")  

        self.optparser.add_option("-l", "--login", 
                                  action="store", type="string", dest="login",
                                  help = "User's UNIX login name")        

        self.optparser.add_option("-p", "--password-hash", 
                                  action="store", type="string", dest="passwd", 
                                  default = "!",
                                  help = "Password hash (default is disabled password)")
        
        self.optparser.add_option("-s", "--ssh-pubkey", 
                                  action="store", type="string", dest="ssh", 
                                  help = "Public SSH key to be added to the new user's authorized_keys file.")        

        self.optparser.add_option("-a", "--admin", 
                                  action="store_true", dest="admin", 
                                  help = "Gives the user sudo privileges.")
        
        self.optparser.add_option("-c", "--certificate", 
                                  action="store", type="string", dest="certificate", 
                                  default = "generated",                                  
                                  help = "Type of certificate. Can be \"generated\" or \"none\" (default is " 
                                         "to generate a certificate for the user)")

                
    def run(self):    
        t_start = time.time()        
        self.parse_options()
                
        if len(self.args) != 2:
            print "You must specify an instance id."
            print "For example: %s [options] gpi-37a8bf17" % self.name
            return 1
          
        inst_id = self.args[1]

        api = API(self.opt.dir)
        (status_code, message, topology_json) = api.instance(inst_id)
        
        if status_code != API.STATUS_SUCCESS:
            self._print_error("Could not access instance.", message)
            return 1 
        else:
            t = Topology.from_json_string(topology_json)
            
            if not t.domains.has_key(self.opt.domain):
                self._print_error("Could not add user", "Domain '%s' does not exist" % self.opt.domain)
                return 1
            
            domain = t.domains[self.opt.domain]
            
            user = User()
            user.set_property("id", self.opt.login)
            user.set_property("password_hash", self.opt.passwd)
            user.set_property("ssh_pkey", self.opt.ssh)
            if self.opt.admin != None:
                user.set_property("admin", self.opt.admin)
            else:
                user.set_property("admin", False)
            user.set_property("certificate", self.opt.certificate)

            domain.add_to_array("users", user)
            
            topology_json = t.to_json_string()

            print "Adding new user to",
            print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",
            status_code, message = api.instance_update(inst_id, topology_json, [], [])
            
            if status_code == API.STATUS_SUCCESS:
                print Fore.GREEN + Style.BRIGHT + "done!"

                self._set_last_gpi(inst_id)
                
                t_end = time.time()
                
                delta = t_end - t_start
                minutes = int(delta / 60)
                seconds = int(delta - (minutes * 60))
                print "Added user in " + Fore.WHITE + Style.BRIGHT + "%i minutes and %s seconds" % (minutes, seconds)
                return 0
            elif status_code == API.STATUS_FAIL:
                self._print_error("Could not update topology.", message)
                return 1
        
        
def gp_instance_add_host_func():
    return gp_instance_add_host(sys.argv).run()     
        
class gp_instance_add_host(Command):
    """
    Adds a new host to a running instance.
    
    The instance identifier must be specified after all other parameters. For example::
    
        gp-instance-add-host --domain simple --id simple-gridftp --run-list role[domain-gridftp-default] gpi-12345678
        
    """
        
    name = "gp-instance-add-host"
    
    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)

        self.optparser.add_option("-m", "--domain", 
                                  action="store", type="string", dest="domain",
                                  help = "Domain that the user will be added to.")  

        self.optparser.add_option("-n", "--id", 
                                  action="store", type="string", dest="id",
                                  help = "Unique identifier for the new host.")        

        self.optparser.add_option("-p", "--depends", 
                                  action="store", type="string", dest="depends", 
                                  default = None,
                                  help = "ID of the node (if any) that the new host depends on.")
        
        self.optparser.add_option("-r", "--run-list", 
                                  action="store", type="string", dest="runlist", 
                                  help = "List of Chef recipes or roles to apply to this host.")     
                
    def run(self):    
        t_start = time.time()        
        self.parse_options()
                
        if len(self.args) != 2:
            print "You must specify an instance id."
            print "For example: %s [options] gpi-37a8bf17" % self.name
            return 1
        
        inst_id = self.args[1]

        api = API(self.opt.dir)
        (status_code, message, topology_json) = api.instance(inst_id)
        
        if status_code != API.STATUS_SUCCESS:
            self._print_error("Could not access instance.", message)
            return 1
        else:
            t = Topology.from_json_string(topology_json)
            
            if not t.domains.has_key(self.opt.domain):
                self._print_error("Could not add host", "Domain '%s' does not exist" % self.opt.domain)
                return 1
            
            domain = t.domains[self.opt.domain]
            
            node = Node()
            node.set_property("id", self.opt.id)
            node.set_property("run_list", self.opt.runlist.split(","))
            if self.opt.depends != None:
                node.set_property("depends", "node:%s" % self.opt.depends)

            domain.add_to_array("nodes", (node))
            
            topology_json = t.to_json_string()

            print "Adding new host to",
            print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",
            status_code, message = api.instance_update(inst_id, topology_json, [], [])
            
            if status_code == API.STATUS_SUCCESS:
                print Fore.GREEN + Style.BRIGHT + "done!"
                
                self._set_last_gpi(inst_id)
                
                t_end = time.time()
                
                delta = t_end - t_start
                minutes = int(delta / 60)
                seconds = int(delta - (minutes * 60))
                print "Added host in " + Fore.WHITE + Style.BRIGHT + "%i minutes and %s seconds" % (minutes, seconds)
                return 0
            elif status_code == API.STATUS_FAIL:
                self._print_error("Could not update topology.", message)
                return 1
        
        
def gp_instance_remove_users_func():
    return gp_instance_remove_users(sys.argv).run()     

class gp_instance_remove_users(Command):
    """
    Removes users from a running instance.
    
    The logins of the users to be removed must be specified after the instance identifier which,
    in turn, must be specified after all other parameters. For example::
    
        gp-instance-remove-users --domain simple gpi-12345678 johnsmith sarahjane
        
    """
        
    name = "gp-instance-remove-users"
    
    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)
                
        self.optparser.add_option("-m", "--domain", 
                                  action="store", type="string", dest="domain",
                                  help = "Removes users from this domain")  
                
    def run(self):    
        t_start = time.time()        
        self.parse_options()
                
        if len(self.args) <= 2:
            print "You must specify an instance id and at least one host."
            print "For example: %s [options] gpi-37a8bf17 simple-wn3" % self.name
            return 1
        
        inst_id = self.args[1]
        users = self.args[2:]
        
        api = API(self.opt.dir)
        (status_code, message, topology_json) = api.instance(inst_id)
        
        if status_code != API.STATUS_SUCCESS:
            self._print_error("Could not access instance.", message)
            return 1
        else:
            t = Topology.from_json_string(topology_json)
            
            if not t.domains.has_key(self.opt.domain):
                self._print_error("Could not remove users", "Domain '%s' does not exist" % self.opt.domain)
                return 1
                            
            removed = []
            domain_users = t.domains[self.opt.domain].users
            
            for user in users:
                if user in domain_users:
                    domain_users.pop(user)
                    removed.append(user)
                        
            remaining = set(removed) ^ set(users)
            for r in remaining:
                print Fore.YELLOW + Style.BRIGHT + "Warning" + Fore.RESET + Style.RESET_ALL + ":",
                print "User %s does not exist." % r

            topology_json = t.to_json_string()

            if len(removed) > 0:
                print "Removing users %s from" % list(removed),
                print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",
                status_code, message = api.instance_update(inst_id, topology_json, [], [])
    
                if status_code == API.STATUS_SUCCESS:
                    print Fore.GREEN + Style.BRIGHT + "done!"
                    
                    self._set_last_gpi(inst_id)
                    
                    t_end = time.time()
                    
                    delta = t_end - t_start
                    minutes = int(delta / 60)
                    seconds = int(delta - (minutes * 60))
                    print "Removed users in " + Fore.WHITE + Style.BRIGHT + "%i minutes and %s seconds" % (minutes, seconds)
                    return 0
                elif status_code == API.STATUS_FAIL:
                    self._print_error("Could not update topology.", message)
                    return 1

              
def gp_instance_remove_hosts_func():
    return gp_instance_remove_hosts(sys.argv).run()             
        
class gp_instance_remove_hosts(Command):
    """
    Removes hosts from a running instance.
    
    The host identifiers must be specified after the instance identifier which,
    in turn, must be specified after all other parameters. For example::
    
        gp-instance-remove-hosts --domain simple gpi-12345678 simple-gridftp simple-condor
        
    """
        
    name = "gp-instance-remove-hosts"
    
    def __init__(self, argv, disable_sigintwatch=False):
        Command.__init__(self, argv, disable_sigintwatch=disable_sigintwatch)

        self.optparser.add_option("-m", "--domain", 
                                  action="store", type="string", dest="domain",
                                  help = "Remove hosts from this domain")  
                
    def run(self):    
        t_start = time.time()        
        self.parse_options()
                
        if len(self.args) <= 2:
            print "You must specify an instance id and at least one host."
            print "For example: %s [options] gpi-37a8bf17 simple-wn3" % self.name
            return 1
        
        inst_id = self.args[1]
        hosts = self.args[2:]
        
        api = API(self.opt.dir)
        (status_code, message, topology_json) = api.instance(inst_id)
        
        if status_code != API.STATUS_SUCCESS:
            self._print_error("Could not access instance.", message)
            return 1 
        else:
            t = Topology.from_json_string(topology_json)
            
            if not t.domains.has_key(self.opt.domain):
                self._print_error("Could not remove hosts", "Domain '%s' does not exist" % self.opt.domain)
                return 1
                            
            removed = []
            nodes = t.domains[self.opt.domain].nodes
            
            for host in hosts:
                if host in nodes:
                    nodes.pop(host)
                    removed.append(host)
                        
            remaining = set(removed) ^ set(hosts)
            for r in remaining:
                print Fore.YELLOW + Style.BRIGHT + "Warning" + Fore.RESET + Style.RESET_ALL + ":",
                print "Host %s does not exist." % r

            topology_json = t.to_json_string()

            if len(removed) > 0:
                print "Removing hosts %s from" % list(removed),
                print Fore.WHITE + Style.BRIGHT + inst_id + Fore.RESET + Style.RESET_ALL + "...",
                status_code, message = api.instance_update(inst_id, topology_json, [], [])
    
                if status_code == API.STATUS_SUCCESS:
                    print Fore.GREEN + Style.BRIGHT + "done!"
                    
                    self._set_last_gpi(inst_id)
                    
                    t_end = time.time()
                    
                    delta = t_end - t_start
                    minutes = int(delta / 60)
                    seconds = int(delta - (minutes * 60))
                    print "Removed hosts in " + Fore.WHITE + Style.BRIGHT + "%i minutes and %s seconds" % (minutes, seconds)
                elif status_code == API.STATUS_FAIL:
                    self._print_error("Could not update topology.", message)
                    return 1
                


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Topology

This module contains classes used to manage a topology. All of these classes
are PersistentObjects, so the properties are defined separately from
the classes themselves, which can make the code a bit hard to read. 
Before diving into this code, you may want to check out the topology 
documentation (in the Globus Provision Documentation), which is automatically
generated from the code in this module.  

"""

from globus.provision.common.persistence import PersistentObject, PropertyTypes,\
    Property
    
class Topology(PersistentObject):
    STATE_NEW = 1
    STATE_STARTING = 2
    STATE_CONFIGURING = 3
    STATE_RUNNING = 4
    STATE_STOPPING = 5
    STATE_STOPPED = 6
    STATE_RESUMING = 7
    STATE_TERMINATING = 8
    STATE_TERMINATED = 9
    STATE_FAILED = 10
    
    # String representation of states
    state_str = {STATE_NEW : "New",
                 STATE_STARTING : "Starting",
                 STATE_CONFIGURING : "Configuring",
                 STATE_RUNNING : "Running",
                 STATE_STOPPING : "Stopping",
                 STATE_STOPPED : "Stopped",
                 STATE_RESUMING : "Resuming",
                 STATE_TERMINATING : "Terminating",
                 STATE_TERMINATED : "Terminated",
                 STATE_FAILED : "Failed"}        
    
    def get_nodes(self):
        nodes = []
        for domain in self.domains.values():
            nodes += [n for n in domain.get_nodes()]
        return nodes       
    
    def get_users(self):
        users = []
        for domain in self.domains.values():
            users += domain.get_users()
        return users    
    
    def gen_hosts_file(self, filename):
        hosts = """127.0.0.1    localhost

# The following lines are desirable for IPv6 capable hosts
::1     localhost ip6-localhost ip6-loopback
fe00::0 ip6-localnet
ff00::0 ip6-mcastprefix
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
ff02::3 ip6-allhosts

"""
        
        nodes = self.get_nodes()
        for n in nodes:
            hosts += " ".join((n.ip, n.hostname, n.hostname.split(".")[0], "\n"))
        
        hostsfile = open(filename, "w")
        hostsfile.write(hosts)
        hostsfile.close()         
        
    def gen_chef_ruby_file(self, filename):
        
        def gen_topology_line(server_name, domain_id, recipes):
            server = domain.find_with_recipes(recipes)
            if len(server) > 0:
                server_node = server[0]
                if len(server) > 1:
                    # TODO: Print a warning saying more than one NFS server has been found
                    pass
                hostname_line = "default[:topology][:domains][\"%s\"][:%s] = \"%s\"\n" % (domain_id, server_name, server_node.hostname)
                ip_line       = "default[:topology][:domains][\"%s\"][:%s_ip] = \"%s\"\n" % (domain_id, server_name, server_node.ip)
                
                return hostname_line + ip_line
            else:
                return ""                              
        
        topology = "default[:topology] = %s\n" % self.to_ruby_hash_string()

        for domain in self.domains.values():
            topology += gen_topology_line("nfs_server", domain.id, ["recipe[provision::nfs_server]", "role[domain-nfsnis]"])
            topology += gen_topology_line("nis_server", domain.id, ["recipe[provision::nis_server]", "role[domain-nfsnis]"])
            topology += gen_topology_line("myproxy_server", domain.id, ["recipe[globus::myproxy]"])
            topology += gen_topology_line("lrm_head", domain.id, ["recipe[condor::condor_head]", "role[domain-condor]"])
        
        topologyfile = open(filename, "w")
        topologyfile.write(topology)
        topologyfile.close()        
    
    def get_depends(self, node):
        if not hasattr(node, "depends"):
            return None
        else:
            return self.get_node_by_id(node.depends[5:])
        
    def get_launch_order(self, nodes):
        order = []
        parents = [n for n in nodes if self.get_depends(n) == None or self.get_depends(n) not in nodes]
        while len(parents) > 0:
            order.append(parents)
            parents = [n for n in nodes if self.get_depends(n) in parents]   
        return order        
    
    def get_node_by_id(self, node_id):
        nodes = self.get_nodes()
        node = [n for n in nodes if n.id == node_id]
        if len(node) == 1:
            return node[0]
        else:
            return None    
        
    def get_deploy_data(self, node, deployer, p_name):
        if node.has_property("deploy_data") and node.deploy_data.has_property(deployer):
            deploy_data = node.deploy_data.get_property(deployer)
            if deploy_data.has_property(p_name):
                return deploy_data.get_property(p_name)
        
        # If node doesn't have requested deploy data, return default (if any)
        if self.has_property("default_deploy_data") and self.default_deploy_data.has_property(deployer):
            deploy_data = self.default_deploy_data.get_property(deployer)
            if deploy_data.has_property(p_name):
                return deploy_data.get_property(p_name)
            
        return None
    
    def get_go_endpoints(self):    
        eps = []
        for domain_name, domain in self.domains.items():
            if domain.has_property("go_endpoints"):
                eps += domain.go_endpoints
        return eps   
    
    def add_domain(self, domain):
        self.add_to_array("domains", domain)
        
    
    
class Domain(PersistentObject):

    def get_nodes(self):
        return self.nodes.values()
    
    def get_users(self):
        return self.users.values()     
    
    def find_with_recipes(self, recipes):
        nodes = []
        for node in self.nodes.values():
            for r in recipes:
                if r in node.run_list:
                    nodes.append(node)
                    continue
        return nodes

    def add_user(self, user):
        self.add_to_array("users", user)
        
    def add_node(self, node):
        self.add_to_array("nodes", node)        

class DeployData(PersistentObject):
    pass

class EC2DeployData(PersistentObject):
    pass

class Node(PersistentObject):
    STATE_NEW = 0
    STATE_STARTING = 1
    STATE_RUNNING_UNCONFIGURED = 2
    STATE_CONFIGURING = 3
    STATE_RUNNING = 4
    STATE_RECONFIGURING = 11
    STATE_STOPPING = 5
    STATE_STOPPING_CONFIGURING = 12
    STATE_STOPPING_CONFIGURED = 13
    STATE_STOPPED = 6
    STATE_RESUMING = 7
    STATE_RESUMED_UNCONFIGURED = 14
    STATE_RESUMED_RECONFIGURING = 15
    STATE_TERMINATING = 8
    STATE_TERMINATED = 9
    STATE_FAILED = 10
    
    # String representation of states
    state_str = {STATE_NEW : "New",
                 STATE_STARTING : "Starting",
                 STATE_RUNNING_UNCONFIGURED : "Running (unconfigured)",
                 STATE_CONFIGURING : "Configuring",
                 STATE_RUNNING : "Running",
                 STATE_RECONFIGURING : "Running (reconfiguring)",
                 STATE_STOPPING : "Stopping",
                 STATE_STOPPING_CONFIGURING : "Stopping (configuring)",
                 STATE_STOPPING_CONFIGURED : "Stopping (configured)",
                 STATE_STOPPED : "Stopped",
                 STATE_RESUMING : "Resuming",
                 STATE_RESUMED_UNCONFIGURED : "Resumed (unconfigured)",
                 STATE_RESUMED_RECONFIGURING : "Resumed (reconfiguring)",
                 STATE_TERMINATING : "Terminating",
                 STATE_TERMINATED : "Terminated",
                 STATE_FAILED : "Failed"}   


class User(PersistentObject):
    pass

class GridMapEntry(PersistentObject):
    pass

class GOEndpoint(PersistentObject):
    pass

Topology.properties = { 
                       "id":
                       Property(name="id",
                                proptype = PropertyTypes.STRING,
                                required = False,
                                description = """
                                Once an instance with this topology has been created,
                                this property will contain the instance ID 
                                (e.g., ``gpi-12345678``) assigned by Globus Provision.
                                """),       
                       
                       "state":
                       Property(name="state",
                                proptype = PropertyTypes.INTEGER,
                                required = False,
                                description = """
                                Once an instance has been created with this topology,
                                this property indicates the state the instance is in.
                                Possible values are:
                                
                                %s
                                
                                """ % 
"\n                                ".join(["* %i: %s" % (i, Topology.state_str[i]) for i in sorted(Topology.state_str.keys())])),     
                                                                        
                       "domains": 
                       Property(name = "domains",
                                proptype = PropertyTypes.ARRAY,
                                items = Domain,
                                items_unique = True,
                                editable = True,
                                required = True,
                                description = """
                                The domains in this topology.
                                """),

                       "default_deploy_data":
                       Property(name = "default_deploy_data",
                                proptype = DeployData,
                                required = False,
                                editable = True,                                
                                description = """
                                The default deployment-specific data for this instance.
                                Individual nodes can override the default values in their
                                ``deploy_data`` property.
                                """)          
                       }

DeployData.properties = { "ec2":
                            Property(name = "ec2",
                                     proptype = EC2DeployData,
                                     required = False,
                                     editable = True,
                                     description = """
                                     Used to specify EC2-specific deployment data.
                                     """)          
                       }

EC2DeployData.properties = { 
                            "instance_type":
                                Property(name = "instance_type",
                                         proptype = PropertyTypes.STRING,
                                         required = False,
                                         editable = True,
                                         description = """
                                         An EC2 instance type (e.g., ``t1.micro``, ``m1.small``, etc.)
                                         """),
                            
                            "instance_id":
                                Property(name = "instance_id",
                                         proptype = PropertyTypes.STRING,
                                         required = False,
                                         description = """
                                         Once a host has been deployed on EC2,
                                         this property will contain its EC2 instance identifier.                                        
                                         """),               
                            "ami":
                                Property(name = "ami",
                                         proptype = PropertyTypes.STRING,
                                         required = False,
                                         editable = True,
                                         description = """
                                         The Amazon Machine Image (AMI) to use when creating
                                         new hosts on Amazon EC2.
                                         """),
                            
                            "security_groups":
                                Property(name = "security_groups",
                                         proptype = PropertyTypes.ARRAY,
                                         items = PropertyTypes.STRING,
                                         items_unique = True,                                         
                                         required = False,
                                         editable = True,
                                         description = """
                                         A list of `Security Groups <http://docs.amazonwebservices.com/AWSEC2/latest/UserGuide/index.html?using-network-security.html>`_
                                         to apply to hosts on EC2. If no security groups are specified,
                                         Globus Provision will create one called ``globus-provision``
                                         that opens the TCP/UDP ports for SSH, GridFTP, and MyProxy. 
                                         """)                                
                       }


Domain.properties = {
                     "id":
                     Property(name="id",
                              proptype = PropertyTypes.STRING,
                              required = True,
                              description = """
                              A unique name for the domain.
                              """),               
                              
                     "nodes":
                     Property(name="nodes",
                              proptype = PropertyTypes.ARRAY,
                              items = Node,
                              items_unique = True,
                              required = True,
                              editable = True,
                              description = """
                              The list of hosts (or *nodes*) in this domain.
                              """),
                              
                     "go_endpoints":                    
                     Property(name="go_endpoints",
                              proptype = PropertyTypes.ARRAY,
                              items = GOEndpoint,
                              required = False,
                              editable = True,
                              description = """
                              The list of Globus Online endpoints defined for this domain.
                              """),
                                    
                     "users": 
                     Property(name="users",
                              proptype = PropertyTypes.ARRAY,
                              items = User,
                              items_unique = True,
                              required = True,
                              editable = True,
                              description = """
                              The list of users in this domain.
                              """),
                            
                     "gridmap":                                           
                     Property(name="gridmap",
                              proptype = PropertyTypes.ARRAY,
                              items = GridMapEntry,
                              required = False,
                              editable = True,
                              description = """
                              The list of gridmap entries for this domain. This
                              is the gridmap that Globus services running on this
                              domain will use to determine if a given user is
                              authorized to access the service.
                              """),
                     }

Node.properties = {
                   "id":
                   Property(name="id",
                            proptype = PropertyTypes.STRING,
                            required = True,
                            description = """
                            A unique identifier for this host. The value of this
                            property is only used for identification purposes
                            (e.g., when printing the status of an instance with
                            ``gp-instance-describe``), and will not affect other
                            properties, like its hostname, etc. (except when using
                            the ``dummy`` deployer).
                            """),
                   "state":
                   Property(name="state",
                            proptype = PropertyTypes.INTEGER,
                            required = False,
                            editable = False,
                            description = """
                            Once an instance with this topology has been created,
                            this property will indicate the state of this particular
                            host.
                            
                            Possible values are:
                                
                                %s
                                
                                """ % 
"\n                                ".join(["* %i: %s" % (i, Node.state_str[i]) for i in sorted(Node.state_str.keys())])),     
                                                                        
                   "run_list":
                   Property(name="run_list",
                            proptype = PropertyTypes.ARRAY,
                            items = PropertyTypes.STRING,
                            required = True,
                            editable = True,
                            description = """
                            The list of Chef recipes to run on this node.
                            See :ref:`sec_runlist` for more details.
                            """),
                            
                   "depends":
                   Property(name="depends",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            editable = True,
                            description = """
                            Sometimes, a host cannot be configured until another host
                            in the topology is configured. For example, NFS clients cannot
                            start until the NFS server is starting. This property is
                            used to specify such dependencies. The value of this property
                            must be of the form node:*node_id*, where *node_id* is
                            the identifier of another node in the domain.
                            
                            For example, if this node depends on ``simple-nfs`` the value
                            of this property would be ``node:simple-nfs``.
                            """),
                            
                   "hostname":
                   Property(name="hostname",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            description = """
                            The fully-qualified hostname assigned by the deployer.
                            """),
                            
                   "ip":
                   Property(name="ip",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            description = """
                            The IP address assigned by the deployer.
                            """),
                            
                   "public_ip":
                   Property(name="public_ip",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            description = """
                            If the IP address assigned in ``ip`` is a private, non-routable,
                            IP address, but the host is also assigned a public IP, it will
                            be assigned to this property by the deployer.
                            """),                            
                            
                   "deploy_data":
                   Property(name = "deploy_data",
                            proptype = DeployData,
                            required = False,
                            description = """
                            Host-specific deployment data. The values specified here
                            will override any values specified in the topology's
                            ``default_deploy__data`` property.
                            """),   
                            
                   "gc_setupkey":
                   Property(name = "gc_setupkey",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            description = """
                            Globus Connect setup key used to obtain a certificate
                            for use by services in this node. 
                            """)                 
                   
                   }          


User.properties = {
                   "id":
                   Property(name="id",                          
                            proptype = PropertyTypes.STRING,
                            required = True,
                            description = """
                            The user's login name.
                            """),
                            
                   "description":
                   Property(name="description",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            editable = True,
                            description = """
                            A description of the user.
                            """),
                            
                   "password_hash":
                   Property(name="password_hash",
                            proptype = PropertyTypes.STRING,
                            required = True,
                            editable = True,
                            description = """
                            The password hash for the user, exactly as it will appear
                            in the shadow file (``/etc/shadow``). To generate a password,
                            you can use the ``mkpasswd`` command. We recommend you generate
                            SHA-512 password hashes. For example::
                            
                                $ mkpasswd -m sha-512 mypassword
                                $6$XrtqyXi4LO$8M/sk6t8zE5Ac.acLPBt577f1eGv.YnUVZPhGmBlQF/YrYnkWQPq7EMfryWEdHm664B.RaY3O8oZtbiQjXfu10
                            
                            The string starting with ``$6$`` is the password hash.
                            
                            You can disable password access for this user by setting this 
                            property to ``!``. 
                            """),
                            
                   "ssh_pkey":
                   Property(name="ssh_pkey",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            editable = True,
                            description = """
                            A public SSH key. If a value is specified for this property,
                            this public key will be added to the user's ``authorized_keys``
                            file.
                            """),
                            
                   "admin":
                   Property(name="admin",
                            proptype = PropertyTypes.BOOLEAN,
                            required = False,
                            editable = True,
                            description = """
                            If ``true``, this user will be granted passwordless sudo
                            access on all hosts in this domain.
                            """),
                            
                   "certificate":
                   Property(name = "certificate",
                            proptype = PropertyTypes.STRING,
                            required = False,
                            description = """
                            This property can take on the following values:
                            
                            * ``"generated"``: A user certificate must be generated for this user.
                            * ``"none"``: Do not generate a certificate for this user.
                            """),           
                   }            

GridMapEntry.properties = {                   
                           "dn": 
                           Property(name="dn",
                                    proptype = PropertyTypes.STRING,
                                    required = True,
                                    description = """
                                    The distinguished name in the gridmap entry
                                    (e.g., ``"/O=Grid/OU=My Grid/CN=J.Random User"``)
                                    """),
                                   
                           "login":
                           Property(name="login",
                                    proptype = PropertyTypes.STRING,
                                    required = True,
                                    editable = True,
                                    description = """
                                    The login the distinguished name will map to.
                                    Must be a valid login in the domain.
                                    """),
                           }       

GOEndpoint.properties = {           
                         
                           "user":
                           Property(name="user",
                                    proptype = PropertyTypes.STRING,
                                    required = True,
                                    description = """
                                    The Globus Online user account in which to create
                                    this endpoint.
                                    """),
                           "name":
                           Property(name="name",
                                    proptype = PropertyTypes.STRING,
                                    required = True,
                                    description = """
                                    The endpoint name.
                                    """),

                           "public":
                           Property(name="public",
                                    proptype = PropertyTypes.BOOLEAN,
                                    required = True,
                                    description = """
                                    Whether the endpoint should be public or not.
                                    """),                         
                                   
                           "gridftp":
                           Property(name="gridftp",
                                    proptype = PropertyTypes.STRING,
                                    required = True,
                                    editable = True,
                                    description = """
                                    The GridFTP server for this endpoint. You can specify
                                    either a fully qualified hostname, or refer to an existing
                                    node in this domain by writing node:*node_id*, 
                                    where *node_id* is the identifier of another node in the domain.
                                    """),
                                   
                           "myproxy":
                           Property(name="myproxy",
                                    proptype = PropertyTypes.STRING,
                                    required = True,
                                    editable = True,
                                    description = """
                                    The MyProxy server that will be used for authenticating
                                    users that want to use this endpoint. You can specify
                                    either a fully qualified hostname, or refer to an existing
                                    node in this domain by writing node:*node_id*, 
                                    where *node_id* is the identifier of another node in the domain.
                                    
                                    Take into account that, to set up this endpoint for
                                    "Globus Online Authentication" (as described in :ref:`sec_go_auth`)
                                    you will need to do the following:
                                    
                                    * Set this property to ``myproxy.globusonline.org``
                                    * For each GO user you want to authorize in this endpoint,
                                      add an entry with the following distinguished name::
                                       
                                           "/C=US/O=Globus Consortium/OU=Globus Connect User/CN=username"
                                      
                                      Where ``username`` is the username of the GO account you
                                      want to authorized. Usually, you will want that DN to map
                                      to the same username in the domain, although this is not
                                      required.

                                    """),
                         
                           "globus_connect_cert":
                           Property(name="globus_connect_cert",
                                    proptype = PropertyTypes.BOOLEAN,
                                    required = False,
                                    description = """
                                    If true, this endpoint will use a Globus Connect certificate.
                                    If not, it will use the host certificate generated by Globus Provision.
                                    Take into account that, for the GridFTP server to be trusted by
                                    Globus Online, it must use a certificate trusted by Globus Online.
                                    Unless you used a CA trusted by Globus Online to generate the certificates
                                    for the topology, you must use a Globus Connect certificate.
                                    """),
                         
                           "globus_connect_cert_dn":
                           Property(name="globus_connect_cert",
                                    proptype = PropertyTypes.STRING,
                                    required = False,
                                    description = """
                                    The DN of the Globus Connect certificate for this endpoint.
                                    """)                                                               
                           
                        }                                                                                                                 



# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
This is the core package of Globus Provision. The core is in charge of managing
instances (creating, starting, stopping, etc.). Although the core orchestrates 
the deployment of instances, it does not have any infrastructure-specific code. 
In other words, the core knows how to start an instance but, when it reaches a 
point where an infrastructure-specific action has to be taken (e.g., "start an 
EC2 instance"), it delegates that task to the appropriate deployer (in the
globus.provision.deploy package). 

The core provides an API that is used by the Command-Line Interface, 
and could potentially be used by other frontends.
"""


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
Persistent objects

A self-documenting, self-validating, persistent object library.
The only persistent backend currently supported is JSON, but
others could be added.
"""

from globus.provision.common.utils import enum

import inspect
import json

class ObjectValidationException(Exception):
    """A simple exception class used for validation exceptions"""
    pass

PropertyTypes = enum("STRING",
                     "INTEGER",
                     "NUMBER",
                     "BOOLEAN",
                     "OBJECT",
                     "ARRAY",
                     "NULL",
                     "ANY")

def pt_to_str(pt, items_type = None):
    if pt == PropertyTypes.STRING:
        return "string"
    elif pt == PropertyTypes.INTEGER:
        return "integer"
    elif pt == PropertyTypes.NUMBER:
        return "number"
    elif pt == PropertyTypes.BOOLEAN:
        return "boolean"
    elif pt == PropertyTypes.OBJECT:
        return "object"
    elif pt == PropertyTypes.ARRAY:
        return "list of %s" % pt_to_str(items_type)
    elif pt == PropertyTypes.NULL:
        return "null"
    elif pt == PropertyTypes.ANY:
        return "any"
    elif inspect.isclass(pt) and issubclass(pt, PersistentObject):
        return pt.__name__
    else:
        return "unknown"

def validate_property_type(value, expected_type, items_type = None, json = False):
    if expected_type == PropertyTypes.STRING:
        valid = isinstance(value, basestring)
    elif expected_type == PropertyTypes.INTEGER:
        valid = isinstance(value, int)
    elif expected_type == PropertyTypes.NUMBER:
        valid = isinstance(value, int) or isinstance(value, float)
    elif expected_type == PropertyTypes.BOOLEAN:
        valid = isinstance(value, bool)
    elif expected_type == PropertyTypes.OBJECT:
        valid = isinstance(value, dict)
    elif expected_type == PropertyTypes.ARRAY:
        if isinstance(value, list):
            valid = True
            for elem in value:
                valid &= validate_property_type(elem, items_type, json = json)
        elif isinstance(value, dict):
            valid = True
            for elem in value.values():
                valid &= validate_property_type(elem, items_type, json = json)
        else:
            valid = False    
    elif expected_type == PropertyTypes.NULL:
        valid = value is None
    elif expected_type == PropertyTypes.ANY:
        valid = True
    elif issubclass(expected_type, PersistentObject):
        # Further validation is done when we convert
        # this object
        if json:
            valid = isinstance(value, dict)
        else:
            valid = isinstance(value, expected_type)
    else:
        valid = False    
    
    return valid

class Property(object):    
    def __init__(self, name, proptype, required, description, editable = False, items = None, items_unique = False):
        self.name = name
        self.type = proptype
        self.required = required
        self.editable = editable
        self.items_unique = items_unique
        self.description = description
        self.items = items
        
        
class PropertyChange(object):
    ADD = 0
    REMOVE = 1
    EDIT = 2    
    
    def __init__(self, change_type):
        self.change_type = change_type


class PrimitivePropertyChange(PropertyChange):

    def __init__(self, change_type, old_value, new_value):
        PropertyChange.__init__(self, change_type)
        self.old_value = old_value
        self.new_value = new_value
        
    def to_dict(self):
        if self.change_type == PropertyChange.ADD:
            return {"__action__": "add"}
        elif self.change_type == PropertyChange.REMOVE:
            return {"__action__": "remove"}
        elif self.change_type == PropertyChange.EDIT:
            d = {}        
            d["old"] = self.old_value
            d["new"] = self.new_value
            return d
        
        
class ArrayPropertyChange(PropertyChange):
    
    def __init__(self, change_type, add, remove, edit):
        PropertyChange.__init__(self, change_type)
        self.add = add
        self.remove = remove
        self.edit = edit
        
    def to_dict(self):
        if self.change_type == PropertyChange.ADD:
            return {"__action__": "add"}
        elif self.change_type == PropertyChange.REMOVE:
            return {"__action__": "remove"}
        elif self.change_type == PropertyChange.EDIT:
            d = {}
        
            if len(self.add) > 0:
                d["ADD"] = self.add
                
            if len(self.remove) > 0:
                d["REMOVE"] = self.remove
            
            if len(self.edit) > 0:
                editd = {}
                for property in self.edit:
                    editd[property] = self.edit[property].to_dict()
                d["EDIT"] = editd
                
            return d
        
        
class ObjectPropertyChange(PropertyChange):
    
    def __init__(self, change_type, changes):
        PropertyChange.__init__(self, change_type)
        self.changes = changes
        
    def to_dict(self):
        if self.change_type == PropertyChange.ADD:
            return {"__action__": "add"}
        elif self.change_type == PropertyChange.REMOVE:
            return {"__action__": "remove"}
        elif self.change_type == PropertyChange.EDIT:
            d = {}
            for property in self.changes:
                d[property] = self.changes[property].to_dict()
            return d
        

class PersistentObject(object):
    def __init__(self):
        self._json_file = None

    def save(self, filename = None):
        if self._json_file == None and filename == None:
            raise Exception("Don't know where to save this topology")
        if filename != None:
            self._json_file = filename
        f = open (self._json_file, "w")
        json_string = self.to_json_string()
        f.write(json_string)
        f.close()
        
    def set_property(self, p_name, p_value):
        # TODO: Validation
        setattr(self, p_name, p_value)
        
    def has_property(self, p_name):
        return hasattr(self, p_name)       
    
    def get_property(self, p_name):
        # TODO: Validation
        return getattr(self, p_name)        
    
    def add_to_array(self, p_name, item_value):
        if not self.properties.has_key(p_name):
            raise ObjectValidationException("%s does not have a %s property" % (type(self).__name__, p_name))
        
        p = self.properties[p_name]
        
        if p.type != PropertyTypes.ARRAY:
            raise ObjectValidationException("Tried to add %s to %s.%s, but it is not an array." % (item_value, type(self).__name__, p_name))
        
        if not validate_property_type(item_value, p.items):
            raise ObjectValidationException("Tried to add %s to %s.%s, but this array contains %s." % (item_value, type(self).__name__, p_name, pt_to_str(p.items)))
        
        if p.items_unique and inspect.isclass(p.items) and issubclass(p.items, PersistentObject):
            if not self.has_property(p_name):
                p_value = {}
                self.set_property(p_name, p_value)
            else:
                p_value = self.get_property(p_name)
                
            setattr(item_value, "parent_%s" % type(self).__name__, self)
            p_value[item_value.id] = item_value
        else:
            if not self.has_property(p_name):
                p_value = []
                self.set_property(p_name, p_value)
            else:
                p_value = self.get_property(p_name)
                
            p_value.append(item_value)
            
    
    def validate_update(self, pobj):
        if type(self) != type(pobj):
            raise ObjectValidationException("Cannot update a %s object with a %s object" % (type(self).__name__, type(pobj).__name__))
        
        changes = {}
        for name, property in self.properties.items():
            self_hasattr = hasattr(self, name)
            pobj_hasattr = hasattr(pobj, name)
            
            if self_hasattr:
                self_value = getattr(self, name)

            if pobj_hasattr:
                pobj_value = getattr(pobj, name)
            
            if property.type in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                if not self_hasattr and pobj_hasattr:
                    if not property.editable:
                        raise ObjectValidationException("Tried to add a property, but it is non-editable (setting '%s' to %s)""" % (name, pobj_value))
                    else:
                        changes[name] = PrimitivePropertyChange(PropertyChange.ADD, None, pobj_value)
                elif self_hasattr and not pobj_hasattr:
                    if not property.editable:
                        raise ObjectValidationException("Tried to remove a property, but it is non-editable (removing '%s' = %s)""" % (name, self_value))
                    else:
                        changes[name] = PrimitivePropertyChange(PropertyChange.REMOVE, self_value, None)
                elif self_hasattr and pobj_hasattr:
                    # If this is a primitive type, check if the value has changed and, if so,
                    # whether the change is allowed.
                    if self_value != pobj_value:
                        if not property.editable:
                            raise ObjectValidationException("Tried to change the value of non-editable property '%s' (from %s to %s)""" % (name, self_value, pobj_value))
                        else:
                            changes[name] = PrimitivePropertyChange(PropertyChange.EDIT, self_value, pobj_value)
            elif property.type == PropertyTypes.ARRAY:
                if not self_hasattr and pobj_hasattr:
                    if not property.editable:
                        raise ObjectValidationException("Tried to add a property, but it is non-editable (setting '%s' to %s)""" % (name, pobj_value))
                    else:
                        changes[name] = ArrayPropertyChange(PropertyChange.ADD, None, None, None)
                elif self_hasattr and not pobj_hasattr:
                    if not property.editable:
                        raise ObjectValidationException("Tried to remove a property, but it is non-editable (removing '%s' = %s)""" % (name, self_value))
                    else:
                        changes[name] = ArrayPropertyChange(PropertyChange.REMOVE, None, None, None)
                elif self_hasattr and pobj_hasattr:
                    if property.items in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                        self_set = set(self_value)
                        pobj_set = set(pobj_value)
                        
                        add = list(pobj_set - self_set)
                        remove = list(self_set - pobj_set)
                        
                        if len(add) + len(remove) > 0:
                            if property.editable: 
                                changes[name] = ArrayPropertyChange(PropertyChange.EDIT, add, remove, {})
                            else:                        
                                raise ObjectValidationException("Tried to add/remove items from non-editable array '%s' (Add: %s  Remove: %s)""" % (name, add, remove))
                    elif inspect.isclass(property.items) and issubclass(property.items, PersistentObject):
                        if property.items_unique:
                            self_set = set(self_value.keys())
                            pobj_set = set(pobj_value.keys())
                            
                            add = list(pobj_set - self_set)
                            remove = list(self_set - pobj_set)
                            
                            if len(add) + len(remove) > 0 and not property.editable:
                                raise ObjectValidationException("Tried to add/remove items from non-editable array '%s' (Add: %s  Remove: %s)""" % (name, add, remove))                            
                            
                            common = list(self_set & pobj_set)
    
                            self_items_value = dict([(k, v) for k, v in self_value.items() if k in common])
                            pobj_items_value = dict([(k, v) for k, v in pobj_value.items() if k in common])
                            
                            edit = {}
                            for s in self_items_value.values():
                                p = pobj_items_value[s.id]
                                item_changes = s.validate_update(p)
                                if len(item_changes.changes) > 0:
                                    if not property.editable:
                                        raise ObjectValidationException("Tried to edit an item in an non-editable array '%s' (Item with id '%s')""" % (name, s.id))
                                    else:
                                        edit[s.id] = item_changes
    
                            if len(add) + len(remove) + len(edit) > 0:
                                changes[name] = ArrayPropertyChange(PropertyChange.EDIT, add, remove, edit)
                        else:
                            # We have no way of telling if individual entries have been edited,
                            # or even if entries have been added/removed, since we don't have
                            # object equality implemented yet.
                            pass
                    elif property.items in (PropertyTypes.ARRAY):
                        raise ObjectValidationException("ARRAYs of ARRAYs not supported.")                            
                    elif property.items in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                        raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")                                                
            elif issubclass(property.type, PersistentObject):
                if not self_hasattr and pobj_hasattr:
                    if not property.editable:
                        raise ObjectValidationException("Tried to add a property, but it is non-editable (setting '%s' to %s)""" % (name, pobj_value))
                    else:
                        changes[name] = ObjectPropertyChange(PropertyChange.ADD, None)
                elif self_hasattr and not pobj_hasattr:
                    if not property.editable:
                        raise ObjectValidationException("Tried to remove a property, but it is non-editable (removing '%s' = %s)""" % (name, self_value))
                    else:
                        changes[name] = ObjectPropertyChange(PropertyChange.REMOVE, None)
                elif self_hasattr and pobj_hasattr:                 
                    property_changes = self_value.validate_update(pobj_value)
                    if len(property_changes.changes) > 0:
                        if not property.editable:
                            raise ObjectValidationException("Tried to to change the value of non-editable property '%s' (Changes: '%s')""" % (name, property_changes.to_dict()))
                        else:
                            changes[name] = property_changes
            elif property.type in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.") 
                
        return ObjectPropertyChange(PropertyChange.EDIT, changes)               

    def to_json_dict(self):
        json = {}
        for name, property in self.properties.items():
            if hasattr(self, name):
                if property.type in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                    value = getattr(self, name)
                elif property.type == PropertyTypes.ARRAY:
                    value = []
                    
                    if inspect.isclass(property.items) and issubclass(property.items, PersistentObject) and property.items_unique:
                        l = getattr(self, name).values()
                    else:
                        l = getattr(self, name)
                    
                    for elem in l:
                        if property.items in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                            value.append(elem)
                        elif issubclass(property.items, PersistentObject):
                            elem_obj = elem.to_json_dict()
                            value.append(elem_obj)
                        elif property.items in (PropertyTypes.ARRAY):
                            raise ObjectValidationException("ARRAYs of ARRAYs not supported.")                            
                        elif property.items in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                            raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")
                elif issubclass(property.type, PersistentObject):
                    value = getattr(self, name).to_json_dict()              
                elif property.type in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                    raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")
                json[name] = value
                
        return json

    def to_json_string(self):
        return json.dumps(self.to_json_dict(), indent=2)

    def __primitive_to_ruby(self, value, p_type):
        if p_type == PropertyTypes.STRING:
            return "\"%s\"" % value
        elif p_type == PropertyTypes.INTEGER:
            return "%i" % value
        elif p_type == PropertyTypes.NUMBER:
            return "%f" % value
        elif p_type == PropertyTypes.BOOLEAN:
            if value == True:
                return "true"
            else:
                return "false" 
        elif p_type == PropertyTypes.NULL:
            return "nil"        

    def to_ruby_hash_string(self):
        hash_str = "{"
        
        obj_items = {}
        for name, property in self.properties.items():
            if hasattr(self, name):
                value = getattr(self, name)
                if property.type in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                    value_str = self.__primitive_to_ruby(value, property.type)
                elif property.type == PropertyTypes.ARRAY and inspect.isclass(property.items) and issubclass(property.items, PersistentObject) and property.items_unique:
                    value_str = "{"
                        
                    items = {}
                    for k, elem in value.items():
                        items[k] = elem.to_ruby_hash_string()
                        
                    value_str += ", ".join([" \"%s\" => %s" % (k,v) for k,v in items.items()])
                    value_str += "}"
                elif property.type == PropertyTypes.ARRAY:
                    value_str = "["
                        
                    items = []
                    for elem in value:
                        if property.items in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                            items.append( self.__primitive_to_ruby(elem, property.items) )
                        elif issubclass(property.items, PersistentObject):
                            items.append( elem.to_ruby_hash_string() )
                        elif property.items in (PropertyTypes.ARRAY):
                            raise ObjectValidationException("ARRAYs of ARRAYs not supported.")                            
                        elif property.items in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                            raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")

                    value_str += ", ".join(items)
                    value_str += "]"
                elif inspect.isclass(property.type) and issubclass(property.type, PersistentObject):
                    value_str = value.to_ruby_hash_string() 
                elif property.type in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                    raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")
                obj_items[name] = value_str        

        hash_str += ", ".join([" :%s => %s" % (k,v) for k,v in obj_items.items()])

        hash_str += "}"
        
        return hash_str

    @classmethod
    def from_json_string(cls, json_string):
        try:
            json_dict = json.loads(json_string)
            return cls.from_json_dict(json_dict)
        except ValueError, ve:
            raise ObjectValidationException("Error parsing JSON. %s" % ve)

    @classmethod
    def from_json_dict(cls, obj_dict):
        obj = cls()
        if not isinstance(obj_dict, dict):
            raise ObjectValidationException("JSON provided for %s is not a dictionary" % cls.__name__)
        
        given_names = set(obj_dict.keys())
        required_names = set([p.name for p in cls.properties.values() if p.required])
        valid_names = set(cls.properties.keys())
        
        # Check whether required fields are present
        missing = required_names - given_names
        if len(missing) > 0:
            raise ObjectValidationException("JSON provided for %s is missing required properties: %s" % (cls.__name__, ", ".join(missing)))
        
        # Check whether there are any unexpected fields
        unexpected = given_names - valid_names
        if len(unexpected) > 0:
            raise ObjectValidationException("Encountered unexpected properties in JSON provided for %s: %s" % (cls.__name__, ", ".join(unexpected)))
        
        for p_name, p_value in obj_dict.items():
            property = cls.properties[p_name]
            if not validate_property_type(p_value, property.type, property.items, json = True):
                raise ObjectValidationException("'%s' is not a valid value for %s.%s. Expected a %s." % (p_value, cls.__name__, p_name, pt_to_str(property.type, property.items)))
            else:
                if property.type in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                    obj.set_property(p_name, p_value)
                    
                elif property.type == PropertyTypes.ARRAY and property.items_unique:
                    if property.items in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                        l = list(set(p_value))
                        if len(l) < len(p_value):
                            raise ObjectValidationException("%s.%s requires unique values, but '%s' contains duplicate values." % (cls.__name__, p_name, p_value))
                        obj.set_property(p_name, l)
                    if inspect.isclass(property.items) and issubclass(property.items, PersistentObject):
                        d = {}
                        for elem in p_value:
                            if not elem.has_key("id"):
                                raise ObjectValidationException("%s.%s requires unique objects, but '%s' does not have an 'id' property." % (cls.__name__, p_name, elem))
                            key = elem["id"]
                            if d.has_key(key):
                                raise ObjectValidationException("%s.%s requires unique objects, but id=%s encountered twice." % (cls.__name__, p_name, key))
                            elem_obj = property.items.from_json_dict(elem)
                            setattr(elem_obj, "parent_%s" % cls.__name__, obj)
                            d[key] = elem_obj
                        obj.set_property(p_name, d)
                    elif property.items in (PropertyTypes.ARRAY):
                        raise ObjectValidationException("ARRAYs of ARRAYs not supported.")                            
                    elif property.items in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                        raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")     
                                           
                elif property.type == PropertyTypes.ARRAY and not property.items_unique:
                        if property.items in (PropertyTypes.STRING, PropertyTypes.INTEGER, PropertyTypes.NUMBER, PropertyTypes.BOOLEAN, PropertyTypes.NULL):
                            l = []
                            for elem in p_value:
                                l.append(elem)
                            obj.set_property(p_name, l)
                        elif inspect.isclass(property.items) and issubclass(property.items, PersistentObject):
                            l = []
                            for elem in p_value:
                                elem_obj = property.items.from_json_dict(elem)
                                l.append(elem_obj)
                            obj.set_property(p_name, l)
                        elif property.items in (PropertyTypes.ARRAY):
                            raise ObjectValidationException("ARRAYs of ARRAYs not supported.")                            
                        elif property.items in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                            raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")
                        
                elif inspect.isclass(property.type) and issubclass(property.type, PersistentObject):
                    p_value_obj = property.type.from_json_dict(p_value)
                    obj.set_property(p_name, p_value_obj)               
                    
                elif property.type in (PropertyTypes.OBJECT, PropertyTypes.ANY):
                    raise ObjectValidationException("Arbitrary types (OBJECT, ANY) not supported.")
                
        return obj



from globus.provision.core.api import API
from globus.provision.tests.common import load_config_file, create_instances_dir,\
    remove_instances_dir, create_temp_config_file
from globus.provision.core.topology import Topology
import json
from globus.provision.cli.api import gp_instance_create, gp_instance_describe,\
    gp_instance_start, gp_instance_stop, gp_instance_list, gp_instance_update,\
    gp_instance_terminate
import os

# TODO: Assumes we're running from the source root directory. Make this location independent.
SAMPLE_DIR = "./samples/"
SAMPLES = ("simple-dummy.conf", "simple-ec2.conf", "single-gridftp-ec2.conf")


def test_samples_api():
    for s in SAMPLES:
        yield check_sample_api, SAMPLE_DIR + s

def test_samples_cli():
    for s in SAMPLES:
        yield check_sample_cli, SAMPLE_DIR + s


def check_sample_api(s):
    instances_dir = create_instances_dir()
    
    config_txt, topology_json = load_config_file(s, dummy = True)
    
    api = API(instances_dir)

    (status_code, message, inst_id) = api.instance_create(topology_json, config_txt)
    assert status_code == API.STATUS_SUCCESS
    check_instance_state(api, inst_id, Topology.STATE_NEW)
    
    (status_code, message, topologies_json) = api.instance_list(None)
    assert status_code == API.STATUS_SUCCESS
    insts = json.loads(topologies_json)
    assert len(insts) == 1
    assert insts[0]["id"] == inst_id
            
    (status_code, message, topologies_json) = api.instance_list([inst_id])
    assert status_code == API.STATUS_SUCCESS, message
    insts = json.loads(topologies_json)
    assert len(insts) == 1
    assert insts[0]["id"] == inst_id
    
    (status_code, message) = api.instance_start(inst_id, [], [])
    assert status_code == API.STATUS_SUCCESS, message
    check_instance_state(api, inst_id, Topology.STATE_RUNNING)

    (status_code, message) = api.instance_stop(inst_id)
    assert status_code == API.STATUS_SUCCESS, message
    check_instance_state(api, inst_id, Topology.STATE_STOPPED)

    (status_code, message) = api.instance_start(inst_id, [], [])
    assert status_code == API.STATUS_SUCCESS, message
    check_instance_state(api, inst_id, Topology.STATE_RUNNING)
    
    (status_code, message) = api.instance_update(inst_id, None, [], [])
    assert status_code == API.STATUS_SUCCESS, message
    check_instance_state(api, inst_id, Topology.STATE_RUNNING)    
    
    (status_code, message) = api.instance_terminate(inst_id)
    assert status_code == API.STATUS_SUCCESS, message
    check_instance_state(api, inst_id, Topology.STATE_TERMINATED)
    
    remove_instances_dir(instances_dir)

def check_sample_cli(s):
    instances_dir = create_instances_dir()
    temp_config_file = instances_dir + "/test.conf"
    create_temp_config_file(s, temp_config_file, dummy = True)

    common_args = ["GPCOMMAND", "-i", instances_dir]

    args = ["-c", temp_config_file]
    rc = gp_instance_create(common_args + args).run()
    assert rc == 0

    insts = [n for n in os.listdir(instances_dir) if n.startswith("gpi-")]
    assert len(insts) == 1
    inst_id = insts[0]
    print inst_id
    args = [inst_id]
    print common_args + args
    rc = gp_instance_describe(common_args + args).run()
    assert rc == 0    

    args = [inst_id, "-v"]
    rc = gp_instance_describe(common_args + args).run()
    assert rc == 0    

    args = [inst_id]
    rc = gp_instance_start(common_args + args, disable_sigintwatch = True).run()
    assert rc == 0    
    
    args = []
    rc = gp_instance_list(common_args + args).run()
    assert rc == 0       

    args = ["-v"]
    rc = gp_instance_list(common_args + args).run()
    assert rc == 0       

    args = [inst_id]
    rc = gp_instance_list(common_args + args).run()
    assert rc == 0       

    args = ["-v", inst_id]
    rc = gp_instance_list(common_args + args).run()
    assert rc == 0       
    return
    args = [inst_id]
    rc = gp_instance_stop(common_args + args, disable_sigintwatch = True).run()
    assert rc == 0       

    args = [inst_id]
    rc = gp_instance_start(common_args + args, disable_sigintwatch = True).run()
    assert rc == 0       

    args = [inst_id]
    rc = gp_instance_update(common_args + args, disable_sigintwatch = True).run()
    assert rc == 0     
    
    args = [inst_id]
    rc = gp_instance_terminate(common_args + args, disable_sigintwatch = True).run()
    assert rc == 0         
    
    remove_instances_dir(instances_dir)


def check_instance_state(api, inst_id, state):
    (status_code, message, topology_json) = api.instance(inst_id)
    assert status_code == API.STATUS_SUCCESS, message
    topology = Topology.from_json_string(topology_json)
    assert topology.state == state
    


# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #
from boto.exception import EC2ResponseError

"""
EC2 images utilities.

Contains code to manage Globus Provision AMIs.
"""

from globus.provision.common.utils import create_ec2_connection
from globus.provision.common.ssh import SSH
from globus.provision.common import log
import time

class EC2AMICreator(object):
    """
    Used to create a Globus Provision AMI.
    """ 
    
    def __init__(self, chef_dir, base_ami, ami_name, instance_type, config):
        self.chef_dir = chef_dir
        self.base_ami = base_ami
        self.ami_name = ami_name
        self.instance_type = instance_type
        self.config = config

        self.keypair = config.get("ec2-keypair")
        self.keyfile = config.get("ec2-keyfile")
        self.hostname = config.get("ec2-server-hostname")
        self.port = config.get("ec2-server-port")
        self.path = config.get("ec2-server-path")
        self.username = config.get("ec2-username")
        self.scratch_dir = config.get("scratch-dir")

    def run(self):
        log.init_logging(2)

        conn = create_ec2_connection(hostname=self.hostname, path=self.path, port=self.port)

        print "Creating instance"
        reservation = conn.run_instances(self.base_ami, 
                                         min_count=1, max_count=1,
                                         instance_type=self.instance_type, 
                                         key_name=self.keypair)
        instance = reservation.instances[0]
        print "Instance %s created. Waiting for it to start..." % instance.id
        
        while True:
            try:
                newstate = instance.update()
                if newstate == "running":
                    break
                time.sleep(2)
            except EC2ResponseError, ec2err:
                if ec2err.error_code == "InvalidInstanceID.NotFound":
                    # If the instance was just created, this is a transient error. 
                    # We just have to wait until the instance appears.
                    pass
                else:
                    raise ec2err            
        
        print "Instance running"
        print self.username, instance.public_dns_name, self.keyfile
        ssh = SSH(self.username, instance.public_dns_name, self.keyfile, None, None)
        try:
            ssh.open()
        except Exception, e:
            print e.message
            exit(1)
        
        print "Copying Chef files"
        ssh.run("sudo mkdir /chef")
        ssh.run("sudo chown -R %s /chef" % self.username)
        ssh.scp_dir("%s" % self.chef_dir, "/chef")
        
        # Some VMs don't include their hostname
        ssh.run("echo \"%s `hostname`\" | sudo tee -a /etc/hosts" % instance.private_ip_address)

        ssh.run("sudo apt-get install lsb-release wget")
        ssh.run("echo \"deb http://apt.opscode.com/ `lsb_release -cs` main\" | sudo tee /etc/apt/sources.list.d/opscode.list")
        ssh.run("wget -qO - http://apt.opscode.com/packages@opscode.com.gpg.key | sudo apt-key add -")
        ssh.run("sudo apt-get update")
        ssh.run("echo 'chef chef/chef_server_url string http://127.0.0.1:4000' | sudo debconf-set-selections")
        ssh.run("sudo apt-get -q=2 install chef")
        
        ssh.run("echo -e \"cookbook_path \\\"/chef/cookbooks\\\"\\nrole_path \\\"/chef/roles\\\"\" > /tmp/chef.conf")        
        ssh.run("echo '{ \"run_list\": \"recipe[provision::ec2]\", \"scratch_dir\": \"%s\" }' > /tmp/chef.json" % self.scratch_dir)

        ssh.run("sudo chef-solo -c /tmp/chef.conf -j /tmp/chef.json")    
        
        ssh.run("sudo update-rc.d -f nis remove")
        ssh.run("sudo update-rc.d -f condor remove")
        ssh.run("sudo update-rc.d -f chef-client remove")
        
        print "Removing private data and authorized keys"
        ssh.run("sudo find /root/.*history /home/*/.*history -exec rm -f {} \;", exception_on_error = False)
        ssh.run("sudo find / -name authorized_keys -exec rm -f {} \;", exception_on_error = False)            
            
        # Apparently instance.stop() will terminate
        # the instance (this is a known bug), so we 
        # use stop_instances instead.
        print "Stopping instance"
        conn.stop_instances([instance.id])
        while instance.update() != "stopped":
            time.sleep(2)
        print "Instance stopped"
        
        print "Creating AMI"
        # Doesn't actually return AMI. Have to make it public manually.
        ami = conn.create_image(instance.id, self.ami_name, description=self.ami_name)       
        
        print "Cleaning up"

        
        print "Terminating instance"
        #conn.terminate_instances([instance.id])
        #while instance.update() != "terminated":
        #    time.sleep(2)
        print "Instance terminated"   

        

class EC2AMIUpdater(object):
    """
    Used to update a Globus Provision AMI.
    """ 
        
    def __init__(self, base_ami, ami_name, files, config):
        self.base_ami = base_ami
        self.ami_name = ami_name
        self.files = files
        
        self.config = config

        self.keypair = config.get("ec2-keypair")
        self.keyfile = config.get("ec2-keyfile")
        self.hostname = config.get("ec2-server-hostname")
        self.port = config.get("ec2-server-port")
        self.path = config.get("ec2-server-path")
        self.username = config.get("ec2-username")

    def run(self):
        log.init_logging(2)
        
        conn = create_ec2_connection(hostname=self.hostname, path=self.path, port=self.port)

        print "Creating instance"
        reservation = conn.run_instances(self.base_ami, 
                                         min_count=1, max_count=1,
                                         instance_type='m1.small', 
                                         key_name=self.keypair)
        instance = reservation.instances[0]
        print "Instance %s created. Waiting for it to start..." % instance.id
        
        while instance.update() != "running":
            time.sleep(2)
        
        print "Instance running."

        print "Opening SSH connection."
        ssh = SSH(self.username, instance.public_dns_name, self.keyfile)
        ssh.open()
        
        print "Copying files"        
        for src, dst in self.files:
            ssh.scp(src, dst)
                  
        print "Removing private data and authorized keys"
        ssh.run("sudo find /root/.*history /home/*/.*history -exec rm -f {} \;", exception_on_error = False)
        ssh.run("sudo find / -name authorized_keys -exec rm -f {} \;", exception_on_error = False)
                  
        # Apparently instance.stop() will terminate
        # the instance (this is a known bug), so we 
        # use stop_instances instead.
        print "Stopping instance"
        conn.stop_instances([instance.id])
        while instance.update() != "stopped":
            time.sleep(2)
        print "Instance stopped"
        
        print "Creating AMI"
        
        # Doesn't actually return AMI. Have to make it public manually.
        ami = conn.create_image(instance.id, self.ami_name, description=self.ami_name)       
        
        if ami != None:
            print ami
        print "Cleaning up"

        
        print "Terminating instance"
        conn.terminate_instances([instance.id])
        while instance.update() != "terminated":
            time.sleep(2)
        print "Instance terminated"   

                



# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #
from globus.provision.core.topology import Topology, Node
import tempfile
from globus.provision.common.threads import SIGINTWatcher

"""
The CLI: A console frontend to Globus Provision that allows a user to request instances, 
start them, etc.
"""

import subprocess
from optparse import OptionParser, OptionGroup
import os
import os.path
import getpass
import colorama
from colorama import Fore, Style
from globus.provision.common import defaults
from globus.provision.common import log
from globus.provision import RELEASE

class Command(object):
    """Base class for all Globus Provisioning commands"""
    
    def __init__(self, argv, root = False, disable_sigintwatch = False):
        
        if root:
            if getpass.getuser() != "root":
                print "Must run as root"
                exit(1)
                
        self.argv = argv
        self.optparser = OptionParser(version = RELEASE)
        self.opt = None
        self.args = None
        
        common_opts = OptionGroup(self.optparser, "Common options", "These options are common to all Globus Provision commands")
        self.optparser.add_option_group(common_opts)
        
        common_opts.add_option("-v", "--verbose", 
                                  action="store_true", dest="verbose", 
                                  help = "Produce verbose output.")

        common_opts.add_option("-d", "--debug", 
                                  action="store_true", dest="debug", 
                                  help = "Write debugging information. Implies -v.")     

        common_opts.add_option("-i", "--instances-dir", 
                                  action="store", type="string", dest="dir", 
                                  default = defaults.INSTANCE_LOCATION,
                                  help = "Use this directory to store information about the instances "
                                         "(instead of the default ~/.globusprovision/instances/)")
        
        colorama.init(autoreset = True)
        
        if not disable_sigintwatch:
            SIGINTWatcher(self.cleanup_after_kill)
                

    def parse_options(self):
        opt, args = self.optparser.parse_args(self.argv)

        self.opt = opt
        self.args = args
        
        if self.opt.debug:
            loglevel = 2
        elif self.opt.verbose:
            loglevel = 1
        else:
            loglevel = 0
            
        log.init_logging(loglevel)
        
    def _run(self, cmd, exit_on_error=True, silent=True):
        if silent:
            devnull = open("/dev/null")
        cmd_list = cmd.split()
        if silent:
            retcode = subprocess.call(cmd_list, stdout=devnull, stderr=devnull)
        else:
            retcode = subprocess.call(cmd_list)
        if silent:
            devnull.close()
        if retcode != 0 and exit_on_error:
            print "Error when running %s" % cmd
            exit(1)        
        return retcode
    
    def _check_exists_file(self, filename):
        if not os.path.exists(filename):
            print "File %s does not exist" % filename
            exit(1)
            
    def _print_error(self, what, reason):
        print colorama.Fore.RED + colorama.Style.BRIGHT + " \033[1;31mERROR\033[0m",
        print ": %s" % what
        print colorama.Fore.WHITE + colorama.Style.BRIGHT + "\033[1;37mReason\033[0m",
        print ": %s" % reason
        
    def _colorize_topology_state(self, state):
        state_str = Topology.state_str[state]
        reset = Fore.RESET + Style.RESET_ALL
        if state == Topology.STATE_NEW:
            return Fore.BLUE + Style.BRIGHT + state_str + reset
        elif state == Topology.STATE_RUNNING:
            return Fore.GREEN + Style.BRIGHT + state_str + reset
        elif state in (Topology.STATE_STARTING, Topology.STATE_CONFIGURING, Topology.STATE_STOPPING, Topology.STATE_RESUMING, Topology.STATE_TERMINATING):
            return Fore.YELLOW + Style.BRIGHT + state_str + reset
        elif state in (Topology.STATE_TERMINATED, Topology.STATE_FAILED):
            return Fore.RED + Style.BRIGHT + state_str + reset
        elif state == Topology.STATE_STOPPED:
            return Fore.MAGENTA + Style.BRIGHT + state_str + reset
        else:
            return state_str

    def _colorize_node_state(self, state):
        state_str = Node.state_str[state]
        reset = Fore.RESET + Style.RESET_ALL
        if state == Node.STATE_NEW:
            return Fore.BLUE + Style.BRIGHT + state_str + reset
        elif state == Node.STATE_RUNNING:
            return Fore.GREEN + Style.BRIGHT + state_str + reset
        elif state in (Node.STATE_STARTING, Node.STATE_RUNNING_UNCONFIGURED, Node.STATE_CONFIGURING, Node.STATE_RECONFIGURING, Node.STATE_STOPPING, Node.STATE_STOPPING_CONFIGURING, Node.STATE_STOPPING_CONFIGURED, Node.STATE_RESUMING, Node.STATE_RESUMED_UNCONFIGURED, Node.STATE_RESUMED_RECONFIGURING, Node.STATE_TERMINATING):
            return Fore.YELLOW + Style.BRIGHT + state_str + reset
        elif state in (Node.STATE_TERMINATED, Node.STATE_FAILED):
            return Fore.RED + Style.BRIGHT + state_str + reset
        elif state == Node.STATE_STOPPED:
            return Fore.MAGENTA + Style.BRIGHT + state_str + reset
        else:
            return state_str
        
    def _pad(self, str, colorstr, width):
        if colorstr == "":
            return str.ljust(width)
        else:
            return colorstr + " " * (width - len(str))        
        
    def _set_last_gpi(self, gpi):
        try:
            uid = os.getuid()
            ppid = os.getppid()
    
            gpi_dir = "%s/.globusprovision-%i" % (tempfile.gettempdir(), uid)
            gpi_file = "%s/%i" % (gpi_dir, ppid) 
            
            if not os.path.exists(gpi_dir):
                os.mkdir(gpi_dir, 0700)
                
            f = open(gpi_file, "w")
            f.write(gpi)
            f.close()
        except Exception, e:
            # Saving the last GPI is just for the benefit of auto-completion.
            # If it doesn't work, worse that will happen is that the user
            # won't be able to autocomplete the last GPI
            pass
        
    def cleanup_after_kill(self):
        print "Globus Provision has been unexpectedly killed and may have left resources"
        print "in an unconfigured state. Use gp-instance-terminate to release resources."
        


from globus.provision.cli import *

#args = []    
#c = demogrid_prepare(args)

#args = ["-n", "a-server"]
#c = demogrid_clone_image(args)

#args = ["--ami", "ami-480df921", 
#        "--keypair", "demogrid", 
#        "--keypair-file", "/home/borja/.ec2/demogrid.pem"]
#c = demogrid_ec2_create_chef_volume(args)

args = ["--ami", "ami-480df921", 
        #"--snapshot", "snap-132adf7e",
        "--name", "DemoGrid Base AMI w Chef files)",
        "--keypair", "demogrid", 
        "--keypair-file", "/home/borja/.ec2/demogrid.pem"]
c = demogrid_ec2_create_ami(args)

#args = []    
#c = demogrid_ec2_launch(args)

c.run()


from globus.provision.common.persistence import *
from globus.provision.core.topology import *
from globus.provision.common.utils import rest_table
import textwrap


def print_section(title, marker):
    print title
    print marker * len(title)
    
def get_ptype(p):
    p_type = pt_to_str(p.type, p.items)
    if inspect.isclass(p.type) and issubclass(p.type, PersistentObject):
        p_type = ":ref:`%s <topology_%s>`" % (p.type.__name__, p.type.__name__)
    elif p.type == PropertyTypes.ARRAY and inspect.isclass(p.items) and issubclass(p.items, PersistentObject):
        p_type = "list of :ref:`%s <topology_%s>`" % (p.items.__name__, p.items.__name__)
    return p_type

def gen_cls_doc(cls):
    print ".. _topology_%s:" %  cls.__name__
    print
    print_section("``%s`` Object" % cls.__name__, "=")
    print
    doc = cls.__doc__
    if doc != None:
        doc = textwrap.dedent(doc).strip()
    else:
        doc = ""    
    
    p_names = cls.properties.keys()
    p_names.sort()
    
    col_names = ["Property Name", "Type", "Required", "Editable"]
    rows = []
    for p_name in p_names:
        p = cls.properties[p_name]
        p_type = get_ptype(p)
            
        row = []
        
        row = [":ref:`%s <topology_%s_%s>`" % (p_name, cls.__name__, p_name), p_type]
        if p.required:
            row.append("**Yes**")
        else:
            row.append("No")

        if p.editable:
            row.append("Yes")
        else:
            row.append("No")

        rows.append(row)
        
    print ".. centered:: Summary of ``%s``'s Properties" % cls.__name__
    print
    print rest_table(col_names, rows)
    
    for p_name in p_names:
        p = cls.properties[p_name]
        description = textwrap.dedent(p.description).strip()

        p_type = get_ptype(p)

        print ".. _topology_%s_%s:" %  (cls.__name__, p_name)
        print     
        print_section(p_name,"-")
        print
        print "*Type*: %s" % p_type
        print
        print "*Required*:",
        if p.required:
            print "**Yes**"
        else:
            print "No"
        print
        print "*Editable*:",
        if p.editable:
            print "Yes"
        else:
            print "No"
        print 
        print description
        print

gen_cls_doc(Topology)
gen_cls_doc(Domain)
gen_cls_doc(Node)
gen_cls_doc(User)
gen_cls_doc(GridMapEntry)
gen_cls_doc(GOEndpoint)
gen_cls_doc(DeployData)
gen_cls_doc(EC2DeployData)





# -------------------------------------------------------------------------- #
# Copyright 2010-2011, University of Chicago                                 #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #
"""
Helper classes for performing Globus Online operations.

"""

import re
import StringIO
import os.path
import paramiko

from pkg_resources import resource_filename

from globus.provision.common.ssh import SSH
from globus.provision.common import log
from globusonline.transfer.api_client import TransferAPIClient, ClientError


class GlobusOnlineException(Exception):
    pass
    
class EndpointExistsException(GlobusOnlineException):
    pass    

class GlobusOnlineHelper(object):
    
    def endpoint_create(self, ep, replace):
            
        if self.endpoint_exists(ep):
            if not replace:
                raise EndpointExistsException, "An endpoint called '%s' already exists. Please choose a different name." % ep.name
            else:
                self.endpoint_remove(ep)                

        gridftp_hostname, gridftp_subject, myproxy_hostname, myproxy_subject = self._get_hostnames_subjects(ep)
            
        self._endpoint_create(ep.name, gridftp_hostname, gridftp_subject, myproxy_hostname, ep.public)       

    def _get_hostnames_subjects(self, ep):
        ca_dn = self.inst.config.get("ca-dn")
        if ca_dn == None:
            ca_dn = "/O=Grid/OU=Globus Provision (generated)"
        else:
            ca_dn = [x.split("=") for x in ca_dn.split(",")]
            ca_dn = "".join(["/%s=%s" % (n.upper().strip(), v.strip()) for n,v in ca_dn])

        if ep.gridftp.startswith("node:"):
            gridftp_node = self.inst.topology.get_node_by_id(ep.gridftp[5:])
            gridftp_hostname = gridftp_node.hostname
            
            if ep.has_property("globus_connect_cert") and ep.globus_connect_cert:
                if ep.has_property("globus_connect_cert_dn"):
                    gridftp_subject = ep.globus_connect_cert_dn
                else:
                    gridftp_subject = None
            else:
                gridftp_subject = "%s/CN=host/%s" % (ca_dn, gridftp_hostname)
        else:
            # Note: If user specifies an arbitrary GridFTP hostname, it
            # will only work if it uses a valid host certificate that
            # is trusted by Globus Online.
            gridftp_hostname = ep.gridftp
            gridftp_subject = None
            
        if ep.myproxy.startswith("node:"):
            myproxy_node = self.inst.topology.get_node_by_id(ep.myproxy[5:])
            myproxy_hostname = myproxy_node.hostname
            myproxy_subject = "%s/CN=host/%s" % (ca_dn, gridftp_hostname)
        else:
            myproxy_hostname = ep.myproxy  
            myproxy_subject = None
            
        return gridftp_hostname, gridftp_subject, myproxy_hostname, myproxy_subject    
    
    @staticmethod
    def from_instance(inst):
        if inst.config.get("go-cert-file") == None and inst.config.get("go-ssh-key") != None:
            # Use SSH
            ssh_key = os.path.expanduser(inst.config.get("go-ssh-key"))
            return GlobusOnlineCLIHelper(inst, ssh_key)
        elif inst.config.get("go-cert-file") != None:
            go_cert_file = os.path.expanduser(inst.config.get("go-cert-file"))
            go_key_file = os.path.expanduser(inst.config.get("go-key-file"))
            go_server_ca = resource_filename("globus.provision", "chef-files/cookbooks/globus/files/default/gd-bundle_ca.cert")
            return GlobusOnlineAPIHelper(inst, go_cert_file, go_key_file, go_server_ca)
        else:
            raise GlobusOnlineException, "Must specify either an SSH key or a certificate to connect to GO"
        

class GlobusOnlineAPIHelper(GlobusOnlineHelper):
    def __init__(self, inst, go_cert_file, go_key_file, go_server_ca):
        self.inst = inst
        self.go_cert_file = go_cert_file
        self.go_key_file = go_key_file
        self.go_server_ca = go_server_ca
        
    def connect(self, username):
        self.api = TransferAPIClient(username, self.go_server_ca, self.go_cert_file, self.go_key_file)

    def endpoint_exists(self, ep):
        try:
            (code, msg, data) = self.api.endpoint(ep.name)

            if code == 200:
                return True
            elif code == 404:
                return False
            else:
                raise GlobusOnlineException, "Unexpected error %i: %s" % (code, msg)            
        except ClientError as ce:
            if ce.status_code == 404:
                return False
            else:
                raise GlobusOnlineException, "Unexpected GO API exception: %s" % ce

    def disconnect(self):
        self.api.close()

    def endpoint_remove(self, ep):
        try:
            (code, msg, data) = self.api.endpoint_delete(ep.name)
            if code >= 400:
                raise GlobusOnlineException, "Unexpected error %i: %s" % (code, msg)            
        except ClientError as ce:
            raise GlobusOnlineException, "Unexpected GO API exception: %s" % ce

    def endpoint_gc_create(self, ep, replace):
        if self.endpoint_exists(ep):
            if not replace:
                raise EndpointExistsException, "An endpoint called '%s' already exists. Please choose a different name." % ep.name
            else:
                self.endpoint_remove(ep)
                   
        try:
            (code, msg, data) = self.api.endpoint_create(ep.name, None, description="Globus Provision endpoint",
                                                         is_globus_connect=True)
            if code >= 400:
                raise GlobusOnlineException, "Unexpected error %i: %s" % (code, msg)            
        except ClientError as ce:
            raise GlobusOnlineException, "Unexpected GO API exception: %s" % ce                   

        gc_setupkey = data["globus_connect_setup_key"]

        return gc_setupkey

    def endpoint_gc_create_finalize(self, ep):
        gridftp_hostname, gridftp_subject, myproxy_hostname, myproxy_subject = self._get_hostnames_subjects(ep)
        
        # This is a kludge until http://jira.globus.org/browse/KOA-1624 gets fixed
        username = self.inst.config.get("ec2-username")
        keyfile = os.path.expanduser(self.inst.config.get("ec2-keyfile"))
        gc_setupkey = self.inst.topology.get_node_by_id(ep.gridftp[5:]).gc_setupkey
        ssh = SSH(username, gridftp_hostname, keyfile, default_outf = None, default_errf = None)
        ssh.open()      
        outf = StringIO.StringIO()
        errf = StringIO.StringIO()
        rc = ssh.run("grid-cert-info -subject -file /etc/grid-security/gc-cert-%s.pem" % (gc_setupkey), outf=outf, errf=errf, exception_on_error=False)
        if rc != 0:
            raise GlobusOnlineException, "Could not create endpoint %s" % ep.name
        ssh.close()

        gridftp_subject = outf.getvalue().strip()     
              
        self.endpoint_remove(ep)
                      
        self._endpoint_create(ep.name, gridftp_hostname, gridftp_subject, myproxy_hostname, ep.public)
        

    def _endpoint_create(self, ep_name, gridftp_hostname, gridftp_subject, myproxy_hostname, public):
        try:
            (code, msg, data) = self.api.endpoint_create(ep_name, gridftp_hostname, description="Globus Provision endpoint",
                                                         scheme="gsiftp", port=2811, subject=gridftp_subject,
                                                         myproxy_server=myproxy_hostname)
            if code >= 400:
                raise GlobusOnlineException, "Unexpected error %i: %s" % (code, msg)            
        except ClientError as ce:
            raise GlobusOnlineException, "Unexpected GO API exception: %s" % ce
   

        if public:
            try:
                (code, msg, data) = self.api.endpoint(ep_name)
                if code >= 400:
                    raise GlobusOnlineException, "Unexpected error %i: %s" % (code, msg)            
            except ClientError as ce:
                raise GlobusOnlineException, "Unexpected GO API exception: %s" % ce    
                  
            data["public"] = True
            
            try:
                (code, msg, data) = self.api.endpoint_update(ep_name, data)
                if code >= 400:
                    raise GlobusOnlineException, "Unexpected error %i: %s" % (code, msg)            
            except ClientError as ce:
                raise GlobusOnlineException, "Unexpected GO API exception: %s" % ce          



class GlobusOnlineCLIHelper(GlobusOnlineHelper):
    def __init__(self, inst, ssh_key):
        self.inst = inst
        self.ssh_key = ssh_key
        
    def connect(self, username):
        self.ssh = SSH(username, "cli.globusonline.org", self.ssh_key, default_outf = None, default_errf = None)
        try:
            self.ssh.open(timeout=10)
        except paramiko.PasswordRequiredException, pre:
            msg = "The specified SSH key (%s) requires a password." % self.ssh_key
            msg += "Please specify a passwordless SSH key."
            raise GlobusOnlineException, msg
        except Exception, e:
            raise GlobusOnlineException, "Could not connect to GO CLI. Cause: %s" % e
        
    def disconnect(self):
        self.ssh.close()

    def endpoint_exists(self, ep):
        rc = self.ssh.run("endpoint-list %s" % (ep.name), exception_on_error=False)
        return (rc == 0)

    def endpoint_remove(self, ep):
        rc = self.ssh.run("endpoint-remove %s" % (ep.name), exception_on_error=False)
        if rc != 0:
            raise GlobusOnlineException, "Could not remove endpoint %s" % ep.name        

    def endpoint_remove_server(self, ep, server):
        rc = self.ssh.run("endpoint-remove %s -p %s" % (ep.name, server), exception_on_error=False)
        if rc != 0:
            raise GlobusOnlineException, "Could not remove server for endpoint %s" % ep.name

    def endpoint_gc_create(self, ep, replace):
        if self.endpoint_exists(ep):
            if not replace:
                raise EndpointExistsException, "An endpoint called '%s' already exists. Please choose a different name." % ep.name
            else:
                self.endpoint_remove(ep)
                   
        outf = StringIO.StringIO()
        errf = StringIO.StringIO()
        
        rc = self.ssh.run("endpoint-add %s --gc" % (ep.name), outf=outf, errf=errf, exception_on_error=False)
        if rc != 0:
            raise GlobusOnlineException, "Could not create endpoint %s" % ep.name
        
        output = outf.getvalue()
        gc_setupkey = re.findall("[0-9abcdef-]+$", output)[0]
        
        return gc_setupkey

    def endpoint_gc_create_finalize(self, ep):
        gridftp_hostname, gridftp_subject, myproxy_hostname, myproxy_subject = self._get_hostnames_subjects(ep)
        
        outf = StringIO.StringIO()
        errf = StringIO.StringIO()
        rc = self.ssh.run("endpoint-list -f subjects  %s" % (ep.name), outf=outf, errf=errf, exception_on_error=False)
        if rc != 0:
            raise GlobusOnlineException, "Could not create endpoint %s" % ep.name

        gridftp_subject = outf.getvalue().strip()      
        ep.set_property("globus_connect_cert_dn", gridftp_subject)
        
        self.endpoint_remove(ep)
           
        self._endpoint_create(ep.name, gridftp_hostname, gridftp_subject, myproxy_hostname, ep.public)

    def endpoint_stop(self, ep):
        gridftp_hostname, gridftp_subject, myproxy_hostname, myproxy_subject = self._get_hostnames_subjects(ep)
        
        self.endpoint_remove_server(ep, gridftp_hostname)        
        
        # Create a disconnected endpoint
        self._endpoint_create(ep.name, "relay-disconnected.globusonline.org", gridftp_subject, "myproxy.globusonline.org", ep.public)

    def endpoint_resume(self, ep):
        gridftp_hostname, gridftp_subject, myproxy_hostname, myproxy_subject = self._get_hostnames_subjects(ep)
        
        # Remove disconnected server
        self.endpoint_remove_server(ep, "relay-disconnected.globusonline.org") 
        
        self._endpoint_create(ep.name, gridftp_hostname, gridftp_subject, myproxy_hostname, ep.public)

    def _endpoint_create(self, ep_name, gridftp_hostname, gridftp_subject, myproxy_hostname, public):   
        rc = self.ssh.run("endpoint-add %s -p %s -s \"%s\"" % (ep_name, gridftp_hostname, gridftp_subject), exception_on_error=False)
        if rc != 0:
            raise GlobusOnlineException, "Could not create endpoint %s" % ep_name

        rc = self.ssh.run("endpoint-modify --myproxy-server=%s %s" % (myproxy_hostname, ep_name), exception_on_error=False)
        if rc != 0:
            raise GlobusOnlineException, "Could not set MyProxy server for endpoint %s" % ep_name

        if public:
            rc = self.ssh.run("endpoint-modify --public %s" % (ep_name), exception_on_error=False)
            if rc != 0:
                raise GlobusOnlineException, "Could not make endpoint %s public" % ep_name






        
        



# -------------------------------------------------------------------------- #
# Copyright 2006-2009, 2011 University of Chicago                            #
# Copyright 2008-2009, Distributed Systems Architecture Group, Universidad   #
# Complutense de Madrid (dsa-research.org)                                   #
#                                                                            #
# Licensed under the Apache License, Version 2.0 (the "License"); you may    #
# not use this file except in compliance with the License. You may obtain    #
# a copy of the License at                                                   #
#                                                                            #
# http://www.apache.org/licenses/LICENSE-2.0                                 #
#                                                                            #
# Unless required by applicable law or agreed to in writing, software        #
# distributed under the License is distributed on an "AS IS" BASIS,          #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
# See the License for the specific language governing permissions and        #
# limitations under the License.                                             #
# -------------------------------------------------------------------------- #

"""
A self-documenting, self-validating configuration file parser
(derived from Python's ConfigParser)
"""

import ConfigParser
import textwrap
import os.path
        
OPTTYPE_INT = 0
OPTTYPE_FLOAT = 1
OPTTYPE_STRING = 2
OPTTYPE_BOOLEAN = 3
OPTTYPE_FILE = 4

class ConfigException(Exception):
    """A simple exception class used for configuration exceptions"""
    pass

class Section(object):
    """
    A section in the configuration file
    """
    
    def __init__(self, name, required, multiple = None, required_if=None, doc=None):
        self.name = name
        self.required = required
        self.required_if = required_if
        self.multiple = multiple
        self.doc = doc
        self.options = {}
        
    def get_doc(self):
        return textwrap.dedent(self.doc).strip()


class Option(object):
    """
    An option in a section
    """
    
    def __init__(self, name, getter, type, required, required_if=None, default=None, valid=None, doc=None):
        self.name = name
        self.getter = getter
        self.type = type
        self.required = required
        self.required_if = required_if
        self.default = default
        self.valid = valid
        self.doc = doc
        
    def get_doc(self):
        return textwrap.dedent(self.doc).strip()

class Config(object):
    """
    A configuration file
    """
    
    def __init__(self, config_file, sections):
        self.config = ConfigParser.ConfigParser()
        self.config.readfp(open(config_file, "r"))        

        self.sections = sections
        self._options = {}
        
        self.__load_all()
        
    def __load_all(self):
        multi_sections = [s for s in self.sections if s.multiple != None]
        required_sections = [s for s in self.sections if s.required]
        conditional_sections = [s for s in self.sections if not s.required and s.required_if != None]
        optional_sections = [s for s in self.sections if not s.required and s.required_if == None]
        
        sections = multi_sections + required_sections + conditional_sections + optional_sections
        
        for sec in sections:
            if sec.multiple != None:
                valid = self.config.get(sec.multiple[0],sec.multiple[1]).split()
                for v in valid:
                    secname = "%s-%s" % (sec.name, v)
                    if self.config.has_section(secname):
                        for opt in sec.options:
                            self.__load_option(sec, secname, opt, True, multiname = v)                        
            else:
                has_section = self.config.has_section(sec.name)
            
                # If the section is required, check if it exists
                if sec.required and not has_section:
                    raise ConfigException, "Required section [%s] not found" % sec.name
                
                # If the section is conditionally required, check that
                # it meets the conditions
                if sec.required_if != None:
                    for req in sec.required_if:
                        (condsec,condopt) = req[0]
                        condvalue = req[1]
                        
                        if self.config.has_option(condsec,condopt) and self.config.get(condsec,condopt) == condvalue:
                            if not has_section:
                                raise ConfigException, "Section '%s' is required when %s.%s==%s" % (sec.name, condsec, condopt, condvalue)
                        
                # Load options
                for opt in sec.options:
                    self.__load_option(sec, sec.name, opt, has_section)

    
    def __load_option(self, sec, secname, opt, has_section, multiname = None):
        # Load a single option
        optname = opt.name
        
        has_option = self.config.has_option(secname, optname)
        
        if not has_option:
            if has_section:
                if opt.required:
                    raise ConfigException, "Required option '%s.%s' not found" % (secname, optname)
                if opt.required_if != None:
                    for req in opt.required_if:
                        (condsec,condopt) = req[0]
                        condvalue = req[1]
                        
                        if self.config.has_option(condsec,condopt) and self.config.get(condsec,condopt) == condvalue:
                            raise ConfigException, "Option '%s.%s' is required when %s.%s==%s" % (secname, optname, condsec, condopt, condvalue)
            
            value = opt.default
            if has_section and opt.type == OPTTYPE_FILE and value != None:
                value = os.path.expanduser(value)
                if not os.path.exists(value):
                    raise ConfigException, "File '%s' does not exist (default for '%s.%s')" % (value, secname, optname)            
        else:
            if opt.type == OPTTYPE_INT:
                value = self.config.getint(secname, optname)
            elif opt.type == OPTTYPE_FLOAT:
                value = self.config.getfloat(secname, optname)
            elif opt.type == OPTTYPE_STRING:
                value = self.config.get(secname, optname)
            elif opt.type == OPTTYPE_BOOLEAN:
                value = self.config.getboolean(secname, optname)
            elif opt.type == OPTTYPE_FILE:
                value = os.path.expanduser(self.config.get(secname, optname))
                if not os.path.exists(value):
                    raise ConfigException, "File '%s' does not exist (specified for '%s.%s')" % (value, secname, optname)
                
            if opt.valid != None:
                if not value in opt.valid:
                    raise ConfigException, "Invalid value '%s' specified for '%s.%s'. Valid values are %s" % (value, secname, optname, opt.valid)
        
        if sec.multiple != None:
            self._options[(multiname,opt.getter)] = value
        else:
            self._options[opt.getter] = value
        
    def get(self, opt):
        return self._options[opt]
    
    def has(self, opt):
        return self._options.has_key(opt)    
        
    @classmethod
    def from_file(cls, configfile):
        file = open (configfile, "r")
        c = ConfigParser.ConfigParser()
        c.readfp(file)
        cfg = cls(c)
        return cfg

        
        